# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import sys
from threading import Thread
from spinn_utilities.log import FormatAdapter
from spinn_utilities.overrides import overrides
from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_front_end_common.abstract_models import (
    AbstractMachineAllocationController)
logger = FormatAdapter(logging.getLogger(__name__))


class MachineAllocationController(
        AbstractMachineAllocationController, metaclass=AbstractBase):
    """ How to manage the allocation of a machine so that it gets cleaned up\
        neatly when the script dies.
    """
    __slots__ = [
        #: boolean flag for telling this thread when the system has ended
        "_exited"
    ]

    def __init__(self, thread_name):
        """
        :param str thread_name:
        """
        thread = Thread(name=thread_name, target=self.__manage_allocation)
        thread.daemon = True
        self._exited = False
        thread.start()

    @overrides(AbstractMachineAllocationController.close)
    def close(self):
        self._exited = True

    @abstractmethod
    def _wait(self):
        """ Wait for some bounded amount of time for a change in the status\
            of the machine allocation.

        :return: Whether the machine is still (believed to be) allocated.
        :rtype: bool
        """

    def _teardown(self):
        """ Perform any extra teardown that the thread requires. Does not\
            need to be overridden if no action is desired."""

    def __manage_allocation(self):
        machine_still_allocated = True
        while machine_still_allocated and not self._exited:
            machine_still_allocated = self._wait()
        self._teardown()
        if not self._exited:
            logger.error(
                "The allocated machine has been released before the end of"
                " the script; this script will now exit")
            sys.exit(1)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_front_end_common.utilities import IOBufExtractor


class ChipIOBufExtractor(object):
    """ Extract the logging output buffers from the machine, and separates\
        lines based on their prefix.
    """

    __slots__ = []

    def __call__(
            self, transceiver, executable_targets, executable_finder,
            app_provenance_file_path=None, system_provenance_file_path=None,
            from_cores="ALL", binary_types=None):
        """
        :param ~spinnman.transceiver.Transceiver transceiver:
        :param ~spinnman.model.ExecutableTargets executable_targets:
        :param ExecutableFinder executable_finder:
        :param app_provenance_file_path:
        :type app_provenance_file_path: str or None
        :param system_provenance_file_path:
        :type system_provenance_file_path: str or None
        :param str from_cores:
        :param str binary_types:
        :return: error_entries, warn_entries
        :rtype: tuple(list(str),list(str))
        """
        extractor = IOBufExtractor(
            transceiver, executable_targets, executable_finder,
            app_provenance_file_path, system_provenance_file_path, from_cores,
            binary_types)
        return extractor.extract_iobuf()



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import time
from spinn_utilities.log import FormatAdapter
from spinnman.messages.scp.enums import Signal
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_front_end_common.utilities.utility_objs import ExecutableType
from spinn_front_end_common.utilities.constants import (
    MICRO_TO_MILLISECOND_CONVERSION)

SAFETY_FINISH_TIME = 0.1

logger = FormatAdapter(logging.getLogger(__name__))


class ApplicationRunner(object):
    """ Ensures all cores are initialised correctly, ran, and completed\
        successfully.
    """

    __slots__ = ["__txrx", "__app_id", "__executable_types", "__syncs"]

    # Wraps up as a PACMAN algorithm
    def __call__(
            self, buffer_manager, notification_interface, executable_types,
            app_id, txrx, runtime, time_scale_factor, no_sync_changes,
            time_threshold, machine, run_until_complete=False):
        """
        :param BufferManager buffer_manager:
        :param NotificationProtocol notification_interface:
        :param executable_types:
        :type executable_types:
            dict(ExecutableType,~spinn_machine.CoreSubsets)
        :param int app_id:
        :param ~spinnman.transceiver.Transceiver txrx:
        :param int runtime:
        :param int time_scale_factor:
        :param int no_sync_changes: Number of synchronisation changes
        :param int time_threshold:
        :param ~spinn_machine.Machine machine:
            the spinn machine instance
        :param bool run_until_complete:
        :return: Number of synchronisation changes
        :rtype: int
        :raises ConfigurationException:
        """
        # pylint: disable=too-many-arguments
        logger.info("*** Running simulation... *** ")

        self.__txrx = txrx
        self.__app_id = app_id
        self.__executable_types = executable_types
        self.__syncs = no_sync_changes

        # wait for all cores to be ready
        self._wait_for_start()

        # set the buffer manager into a resume state, so that if it had ran
        # before it'll work again
        buffer_manager.resume()

        # every thing is in sync0 so load the initial buffers
        buffer_manager.load_initial_buffers()

        # clear away any router diagnostics that have been set due to all
        # loading applications
        for chip in machine.chips:
            if not chip.virtual:
                txrx.clear_router_diagnostic_counters(chip.x, chip.y)

        # wait till external app is ready for us to start if required
        notification_interface.wait_for_confirmation()

        # set off the executables that are in sync state
        # (sending to all is just as safe)
        self._send_sync_signal()

        # Send start notification to external applications
        notification_interface.send_start_resume_notification()

        if runtime is None and not run_until_complete:
            # Do NOT stop the buffer manager at end; app is using it still
            logger.info("Application is set to run forever; exiting")
        else:
            # Wait for the application to finish
            try:
                self._run_wait(
                    run_until_complete, runtime, time_scale_factor,
                    time_threshold)
            finally:
                # Stop the buffer manager after run
                buffer_manager.stop()

            # Send stop notification to external applications
            notification_interface.send_stop_pause_notification()

        return self.__syncs

    def _run_wait(self, run_until_complete, runtime, time_scale_factor,
                  time_threshold):
        """
        :param bool run_until_complete:
        :param int runtime:
        :param int time_scale_factor:
        :param float time_threshold:
        """
        if not run_until_complete:
            factor = time_scale_factor / MICRO_TO_MILLISECOND_CONVERSION
            scaled_runtime = runtime * factor
            time_to_wait = scaled_runtime + SAFETY_FINISH_TIME
            logger.info(
                "Application started; waiting {}s for it to stop",
                time_to_wait)
            time.sleep(time_to_wait)
            self._wait_for_end(timeout=time_threshold)
        else:
            logger.info("Application started; waiting until finished")
            self._wait_for_end()

    def _wait_for_start(self, timeout=None):
        """
        :param timeout:
        :type timeout: float or None
        """
        for executable_type in self.__executable_types:
            self.__txrx.wait_for_cores_to_be_in_state(
                self.__executable_types[executable_type], self.__app_id,
                executable_type.start_state, timeout=timeout)

    def _send_sync_signal(self):
        """ Let apps that use the simulation interface or sync signals \
            commence running their main processing loops. This is done with \
            a very fast synchronisation barrier and a signal.
        """
        if (ExecutableType.USES_SIMULATION_INTERFACE in self.__executable_types
                or ExecutableType.SYNC in self.__executable_types):
            # locate all signals needed to set off executables
            sync_signal = self._determine_simulation_sync_signals()

            # fire all signals as required
            self.__txrx.send_signal(self.__app_id, sync_signal)

    def _wait_for_end(self, timeout=None):
        """
        :param timeout:
        :type timeout: float or None
        """
        for executable_type in self.__executable_types:
            self.__txrx.wait_for_cores_to_be_in_state(
                self.__executable_types[executable_type], self.__app_id,
                executable_type.end_state, timeout=timeout)

    def _determine_simulation_sync_signals(self):
        """ Determines the start states, and creates core subsets of the\
            states for further checks.

        :return: the sync signal
        :rtype: ~.Signal
        :raises ConfigurationException:
        """
        sync_signal = None

        if ExecutableType.USES_SIMULATION_INTERFACE in self.__executable_types:
            if self.__syncs % 2 == 0:
                sync_signal = Signal.SYNC0
            else:
                sync_signal = Signal.SYNC1
            # when it falls out of the running, it'll be in a next sync
            # state, thus update needed
            self.__syncs += 1

        # handle the sync states, but only send once if they work with
        # the simulation interface requirement
        if ExecutableType.SYNC in self.__executable_types:
            if sync_signal == Signal.SYNC1:
                raise ConfigurationException(
                    "There can only be one SYNC signal per run. This is "
                    "because we cannot ensure the cores have not reached the "
                    "next SYNC state before we send the next SYNC. Resulting "
                    "in uncontrolled behaviour")
            sync_signal = Signal.SYNC0

        return sync_signal



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from spinn_utilities.progress_bar import ProgressBar
from spinnman.model.enums import CPUState
from spinn_front_end_common.utilities.utility_objs import ExecutableType
from spinn_front_end_common.utilities.scp import UpdateRuntimeProcess


class ChipRuntimeUpdater(object):
    """ Updates the runtime of an application running on a SpiNNaker machine.
    """

    __slots__ = []

    def __call__(
            self, txrx, app_id, executable_types, run_until_timesteps,
            current_timesteps, n_sync_steps):
        """
        :param ~spinnman.transceiver.Transceiver transceiver:
        :param int app_id:
        :param executable_types:
        :type executable_types: dict(ExecutableType,~spinn_machine.CoreSubsets)
        :param run_until_timesteps:
        :type run_until_timesteps: int or None
        :param int current_timesteps:
        :param n_sync_steps:
        :type n_sync_steps: int or None
        """
        core_subsets = \
            executable_types[ExecutableType.USES_SIMULATION_INTERFACE]

        ready_progress = ProgressBar(
            total_number_of_things_to_do=1,
            string_describing_what_being_progressed=(
                "Waiting for cores to be either in PAUSED or READY state"))
        txrx.wait_for_cores_to_be_in_state(
            core_subsets, app_id, [CPUState.PAUSED, CPUState.READY],
            error_states=frozenset({
                CPUState.RUN_TIME_EXCEPTION, CPUState.WATCHDOG,
                CPUState.FINISHED}))
        ready_progress.end()

        infinite_run = 0
        if run_until_timesteps is None:
            infinite_run = 1
            run_until_timesteps = 0
            current_timesteps = 0

        # TODO: Expose the connection selector in SpiNNMan
        process = UpdateRuntimeProcess(txrx.scamp_connection_selector)
        process.update_runtime(
            current_timesteps, run_until_timesteps, infinite_run, core_subsets,
            len(core_subsets), n_sync_steps)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import math
from spinn_utilities.overrides import overrides
from spalloc import Job
from spalloc.states import JobState
from spinn_front_end_common.abstract_models import (
    AbstractMachineAllocationController)
from spinn_front_end_common.abstract_models.impl import (
    MachineAllocationController)


class _SpallocJobController(MachineAllocationController):
    __slots__ = [
        # the spalloc job object
        "_job",
        # the current job's old state
        "_state"
    ]

    def __init__(self, job):
        """
        :param ~spalloc.job.Job job:
        """
        if job is None:
            raise Exception("must have a real job")
        self._job = job
        self._state = job.state
        super().__init__("SpallocJobController")

    @overrides(AbstractMachineAllocationController.extend_allocation)
    def extend_allocation(self, new_total_run_time):
        # Does Nothing in this allocator - machines are held until exit
        pass

    @overrides(AbstractMachineAllocationController.close)
    def close(self):
        super().close()
        self._job.destroy()

    @property
    def power(self):
        """
        :rtype: bool
        """
        return self._job.power

    def set_power(self, power):
        """
        :param bool power:
        """
        self._job.set_power(power)
        if power:
            self._job.wait_until_ready()

    @overrides(AbstractMachineAllocationController.where_is_machine)
    def where_is_machine(self, chip_x, chip_y):
        """
        :param int chip_x:
        :param int chip_y:
        :rtype: tuple(int,int,int)
        """
        return self._job.where_is_machine(chip_y=chip_y, chip_x=chip_x)

    @overrides(MachineAllocationController._wait)
    def _wait(self):
        try:
            if self._state != JobState.destroyed:
                self._state = self._job.wait_for_state_change(self._state)
        except TypeError:
            pass
        except Exception as e:  # pylint: disable=broad-except
            if not self._exited:
                raise e
        return self._state != JobState.destroyed

    @overrides(MachineAllocationController._teardown)
    def _teardown(self):
        if not self._exited:
            self._job.close()
        super()._teardown()


class SpallocAllocator(object):
    """ Request a machine from a SPALLOC server that will fit the given\
        number of chips.
    """

    # Use a worst case calculation
    _N_CHIPS_PER_BOARD = 48.0
    _MACHINE_VERSION = 5

    def __call__(
            self, spalloc_server, spalloc_user, n_chips=None, n_boards=None,
            spalloc_port=None, spalloc_machine=None):
        """
        :param str spalloc_server:
            The server from which the machine should be requested
        :param str spalloc_user: The user to allocate the machine to
        :param n_chips: The number of chips required.
            IGNORED if n_boards is not None
        :type n_chips: int or None
        :param int n_boards: The number of boards required
        :type n_boards: int or None
        :param int spalloc_port: The optional port number to speak to spalloc
        :param str spalloc_machine: The optional spalloc machine to use
        :rtype: tuple(str, int, None, bool, bool, None, None,
            MachineAllocationController)
        """
        # pylint: disable=too-many-arguments

        # Work out how many boards are needed
        if n_boards is None:
            n_boards = float(n_chips) / self._N_CHIPS_PER_BOARD
            # If the number of boards rounded up is less than 10% of a board
            # bigger than the actual number of boards,
            # add another board just in case.
            if math.ceil(n_boards) - n_boards < 0.1:
                n_boards += 1
            n_boards = int(math.ceil(n_boards))

        spalloc_kw_args = {
            'hostname': spalloc_server,
            'owner': spalloc_user
        }
        if spalloc_port is not None:
            spalloc_kw_args['port'] = spalloc_port
        if spalloc_machine is not None:
            spalloc_kw_args['machine'] = spalloc_machine

        job, hostname = self._launch_job(n_boards, spalloc_kw_args)
        machine_allocation_controller = _SpallocJobController(job)

        return (
            hostname, self._MACHINE_VERSION, None, False,
            False, None, None, machine_allocation_controller
        )

    def _launch_job(self, n_boards, spalloc_kw_args):
        """
        :param int n_boards:
        :param dict(str, str or int) spalloc_kw_args:
        :rtype: tuple(~.Job, str)
        """
        job = Job(n_boards, **spalloc_kw_args)
        try:
            job.wait_until_ready()
            # get param from jobs before starting, so that hanging doesn't
            # occur
            return job, job.hostname
        except Exception:
            job.destroy()
            raise



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import logging
import os
import errno
import shutil
import time
import spinn_utilities.conf_loader as conf_loader
from spinn_utilities.log import FormatAdapter
from spinn_machine import Machine
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_front_end_common.utilities.helpful_functions import (
    read_config, read_config_boolean, read_config_int)

logger = FormatAdapter(logging.getLogger(__name__))

APP_DIRNAME = 'application_generated_data_files'
CONFIG_FILE = "spinnaker.cfg"
FINISHED_FILENAME = "finished"
REPORTS_DIRNAME = "reports"
TIMESTAMP_FILENAME = "time_stamp"
WARNING_LOGS_FILENAME = "warning_logs.txt"


class ConfigHandler(object):
    """ Superclass of AbstractSpinnakerBase that handles function only \
        dependent of the config and the order its methods are called.
    """

    __slots__ = [

        # the interface to the cfg files. supports get get_int etc
        "_config",

        #
        "_json_folder",

        #
        "_provenance_file_path",

        #
        "_app_provenance_file_path",

        #
        "_system_provenance_file_path",

        #
        "_report_default_directory",

        #
        "_report_simulation_top_directory",

        #
        "_this_run_time_string",

        #
        "_use_virtual_board",

        # If not None, path to append pacman executor provenance info to
        "_pacman_executor_provenance_path",
    ]

    def __init__(self, configfile, default_config_paths, validation_cfg):
        """
        :param str configfile:
            The base name of the configuration file(s).
            Should not include any path components.
        :param list(str) default_config_paths:
            The list of files to get default configurations from.
        :type default_config_paths: list(str) or None
        :param validation_cfg:
            The list of files to read a validation configuration from.
            If None, no such validation is performed.
        :type validation_cfg: list(str) or None
        """
        # global params
        if default_config_paths is None:
            default_config_paths = []
        default_config_paths.insert(0, os.path.join(
            os.path.dirname(__file__), CONFIG_FILE))

        self._config = conf_loader.load_config(
            filename=configfile, defaults=default_config_paths,
            validation_cfg=validation_cfg)

        # set up machine targeted data
        self._use_virtual_board = self._config.getboolean(
            "Machine", "virtual_board")

        # Pass max_machine_cores to Machine so if effects everything!
        max_machine_core = self._read_config_int("Machine", "max_machine_core")
        if max_machine_core is not None:
            Machine.set_max_cores_per_chip(max_machine_core)

        self._json_folder = None
        self._provenance_file_path = None
        self._report_default_directory = None
        self._report_simulation_top_directory = None
        self._this_run_time_string = None

    @property
    def config(self):
        return self._config

    def _adjust_config(self, runtime, debug_enable_opts, report_disable_opts):
        """ Adjust and checks config based on runtime and mode

        :param runtime:
        :type runtime: int or bool
        :param frozenset(str) debug_enable_opts:
        :param frozenset(str) report_disable_opts:
        :raises ConfigurationException:
        """
        if self._config.get("Mode", "mode") == "Debug":
            for option in self._config.options("Reports"):
                # options names are all lower without _ inside config
                if option in debug_enable_opts or option[:5] == "write":
                    try:
                        if not self._config.get_bool("Reports", option):
                            self._config.set("Reports", option, "True")
                            logger.info("As mode == \"Debug\", [Reports] {} "
                                        "has been set to True", option)
                    except ValueError:
                        pass
        elif not self._config.getboolean("Reports", "reportsEnabled"):
            for option in self._config.options("Reports"):
                # options names are all lower without _ inside config
                if option in report_disable_opts or option[:5] == "write":
                    try:
                        if not self._config.get_bool("Reports", option):
                            self._config.set("Reports", option, "False")
                            logger.info(
                                "As reportsEnabled == \"False\", [Reports] {} "
                                "has been set to False", option)
                    except ValueError:
                        pass

        if runtime is None:
            if self._config.getboolean(
                    "Reports", "write_energy_report") is True:
                self._config.set("Reports", "write_energy_report", "False")
                logger.info("[Reports]write_energy_report has been set to "
                            "False as runtime is set to forever")
            if self._config.get_bool(
                    "EnergySavings", "turn_off_board_after_discovery") is True:
                self._config.set(
                    "EnergySavings", "turn_off_board_after_discovery", "False")
                logger.info("[EnergySavings]turn_off_board_after_discovery has"
                            " been set to False as runtime is set to forever")

        if self._use_virtual_board:
            if self._config.getboolean(
                    "Reports", "write_energy_report") is True:
                self._config.set("Reports", "write_energy_report", "False")
                logger.info("[Reports]write_energy_report has been set to "
                            "False as using virtual boards")
            if self._config.get_bool(
                    "EnergySavings", "turn_off_board_after_discovery") is True:
                self._config.set(
                    "EnergySavings", "turn_off_board_after_discovery", "False")
                logger.info("[EnergySavings]turn_off_board_after_discovery has"
                            " been set to False as s using virtual boards")
            if self._config.getboolean(
                    "Reports", "write_board_chip_report") is True:
                self._config.set("Reports", "write_board_chip_report", "False")
                logger.info("[Reports]write_board_chip_report has been set to"
                            " False as using virtual boards")

    def child_folder(self, parent, child_name, must_create=False):
        """
        :param str parent:
        :param str child_name:
        :param bool must_create:
            If `True`, the directory named by `child_name` (but not necessarily
            its parents) must be created by this call, and an exception will be
            thrown if this fails.
        :return: The fully qualified name of the child folder.
        :rtype: str
        :raises OSError: if the directory existed ahead of time and creation
            was required by the user
        """
        child = os.path.join(parent, child_name)
        if must_create:
            # Throws OSError or FileExistsError (a subclass of OSError) if the
            # directory exists.
            os.makedirs(child)
        elif not os.path.exists(child):
            self._make_dirs(child)
        return child

    def _remove_excess_folders(self, max_kept, starting_directory):
        try:
            files_in_report_folder = os.listdir(starting_directory)

            # while there's more than the valid max, remove the oldest one
            if len(files_in_report_folder) > max_kept:

                # sort files into time frame
                files_in_report_folder.sort(
                    key=lambda temp_file: os.path.getmtime(
                        os.path.join(starting_directory, temp_file)))

                # remove only the number of files required, and only if they
                # have the finished flag file created
                num_files_to_remove = len(files_in_report_folder) - max_kept
                files_removed = 0
                files_not_closed = 0
                for current_oldest_file in files_in_report_folder:
                    finished_flag = os.path.join(os.path.join(
                        starting_directory, current_oldest_file),
                        FINISHED_FILENAME)
                    if os.path.exists(finished_flag):
                        shutil.rmtree(os.path.join(
                            starting_directory, current_oldest_file),
                            ignore_errors=True)
                        files_removed += 1
                    else:
                        files_not_closed += 1
                    if files_removed + files_not_closed >= num_files_to_remove:
                        break
                if files_not_closed > max_kept // 4:
                    logger.warning(
                        "{} has {} old reports that have not been closed",
                        starting_directory, files_not_closed)
        except IOError:
            # This might happen if there is an open file, or more than one
            # process in the same folder, but we shouldn't die because of it
            pass

    def _set_up_report_specifics(self, n_calls_to_run):
        """
        :param int n_calls_to_run:
            the counter of how many times run has been called.
        """

        default_report_file_path = self._config.get_str(
            "Reports", "default_report_file_path")
        # determine common report folder
        if default_report_file_path == "DEFAULT":
            directory = os.getcwd()

            # global reports folder
            report_default_directory = self.child_folder(
                directory, REPORTS_DIRNAME)
        elif default_report_file_path == "REPORTS":
            report_default_directory = REPORTS_DIRNAME
            if not os.path.exists(report_default_directory):
                self._make_dirs(report_default_directory)
        else:
            report_default_directory = self.child_folder(
                default_report_file_path, REPORTS_DIRNAME)

        # clear and clean out folders considered not useful anymore
        if os.listdir(report_default_directory):
            self._remove_excess_folders(
                self._config.getint("Reports", "max_reports_kept"),
                report_default_directory)

        # determine the time slot for later while also making the report folder
        if self._this_run_time_string is None:
            while True:
                try:
                    timestamp = self.__make_timestamp()
                    self._report_simulation_top_directory = self.child_folder(
                        report_default_directory, timestamp, must_create=True)
                    self._this_run_time_string = timestamp
                    break
                except OSError:
                    time.sleep(0.5)
        else:
            self._report_simulation_top_directory = self.child_folder(
                report_default_directory, self._this_run_time_string)

        # create sub folder within reports for sub runs
        # (where changes need to be recorded)
        self._report_default_directory = self.child_folder(
            self._report_simulation_top_directory, "run_{}".format(
                n_calls_to_run))

        # store timestamp in latest/time_stamp for provenance reasons
        time_of_run_file_name = os.path.join(
            self._report_simulation_top_directory, TIMESTAMP_FILENAME)
        with open(time_of_run_file_name, "w") as f:
            f.writelines(self._this_run_time_string)

        if self._read_config_boolean("Logging",
                                     "warnings_at_end_to_file"):
            log_report_file = os.path.join(
                self._report_default_directory, WARNING_LOGS_FILENAME)
            logger.set_report_File(log_report_file)

    @staticmethod
    def __make_timestamp():
        now = datetime.datetime.now()
        return "{:04}-{:02}-{:02}-{:02}-{:02}-{:02}-{:02}".format(
            now.year, now.month, now.day,
            now.hour, now.minute, now.second, now.microsecond)

    def _set_up_output_folders(self, n_calls_to_run):
        """ Sets up all outgoing folders by creating a new timestamp folder
            for each and clearing

        :param int n_calls_to_run:
            the counter of how many times run has been called.
        :rtype: None
        """

        # set up reports default folder
        self._set_up_report_specifics(n_calls_to_run)

        if self._read_config_boolean("Reports",
                                     "writePacmanExecutorProvenance"):
            self._pacman_executor_provenance_path = os.path.join(
                self._report_default_directory,
                "pacman_executor_provenance.rpt")

        self._json_folder = os.path.join(
            self._report_default_directory, "json_files")
        if not os.path.exists(self._json_folder):
            self._make_dirs(self._json_folder)

        # make a folder for the provenance data storage
        self._provenance_file_path = os.path.join(
            self._report_default_directory, "provenance_data")
        if not os.path.exists(self._provenance_file_path):
            self._make_dirs(self._provenance_file_path)

        # make application folder for provenance data storage
        self._app_provenance_file_path = os.path.join(
            self._provenance_file_path, "app_provenance_data")
        if not os.path.exists(self._app_provenance_file_path):
            self._make_dirs(self._app_provenance_file_path)

        # make system folder for provenance data storage
        self._system_provenance_file_path = os.path.join(
            self._provenance_file_path, "system_provenance_data")
        if not os.path.exists(self._system_provenance_file_path):
            self._make_dirs(self._system_provenance_file_path)

    def write_finished_file(self):
        """ Write a finished file that allows file removal to only remove
            folders that are finished.
        """
        report_file_name = os.path.join(self._report_simulation_top_directory,
                                        FINISHED_FILENAME)
        with open(report_file_name, "w") as f:
            f.writelines("finished")

    def _read_config(self, section, item):
        return read_config(self._config, section, item)

    def _read_config_int(self, section, item):
        return read_config_int(self._config, section, item)

    def _read_config_boolean(self, section, item):
        return read_config_boolean(self._config, section, item)

    @property
    def machine_time_step(self):
        """ The machine timestep, in microseconds

        :rtype: int
        """
        return self._read_config_int("Machine", "machine_time_step")

    @machine_time_step.setter
    def machine_time_step(self, new_value):
        self._config.set("Machine", "machine_time_step", new_value)

    @property
    def time_scale_factor(self):
        """ The time scaling factor.

        :rtype: int
        """
        return self._read_config_int("Machine", "time_scale_factor")

    @time_scale_factor.setter
    def time_scale_factor(self, new_value):
        self._config.set("Machine", "time_scale_factor", new_value)

    def set_up_timings(self, machine_time_step=None, time_scale_factor=None):
        """ Set up timings of the machine

        :param machine_time_step:
            An explicitly specified time step for the machine.  If None,
            the value is read from the config
        :type machine_time_step: int or None
        :param time_scale_factor:
            An explicitly specified time scale factor for the simulation.
            If None, the value is read from the config
        :type time_scale_factor: int or None
        """

        # set up timings
        if machine_time_step is not None:
            self.machine_time_step = machine_time_step

        if self.machine_time_step <= 0:
            raise ConfigurationException(
                "invalid machine_time_step {}: must greater than zero".format(
                    self.machine_time_step))

        if time_scale_factor is not None:
            self.time_scale_factor = time_scale_factor

    @staticmethod
    def _make_dirs(path):
        # Workaround for Python 2/3 Compatibility (Python 3 raises on exists)
        try:
            os.makedirs(path)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise



# Copyright (c) 2019-2020 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import sys
from collections import defaultdict
from spinn_utilities.log import FormatAdapter
from spinn_front_end_common.abstract_models import AbstractHasAssociatedBinary
from .bit_field_summary import BitFieldSummary
from spinn_front_end_common.utilities.utility_objs import (
    ProvenanceDataItem, ExecutableType)

logger = FormatAdapter(logging.getLogger(__name__))
_FILE_NAME = "bit_field_compressed_summary.rpt"
# provenance data item names
PROV_TOP_NAME = "bit_field_router_provenance"
PROV_CHIP_NAME = "router_at_chip_{}_{}"
MERGED_NAME = "bit_fields_merged"
NOT_APPLICABLE = "N/A"


def generate_provenance_item(x, y, bit_fields_merged):
    """
    Generates a provenance item in the format BitFieldCompressorReport expects
    :param x:
    :param y:
    :param bit_fields_merged:
    :return:
    """
    # prov names
    names = list()
    names.append(PROV_TOP_NAME)
    names.append(PROV_CHIP_NAME.format(x, y))
    names.append(MERGED_NAME)
    return ProvenanceDataItem(names, str(bit_fields_merged))


class BitFieldCompressorReport(object):
    """ Generates a report that shows the impact of the compression of \
        bitfields into the routing table.
    """
    def __call__(
            self, report_default_directory, machine_graph, placements,
            provenance_items=None):
        """
        :param str report_default_directory: report folder
        :param ~pacman.model.graphs.machine.MachineGraph machine_graph:
            the machine graph
        :param ~pacman.model.placements.Placements placements: the placements
        :param list(ProvenanceDataItem) provenance_items: prov items
        :type provenance_items: list(ProvenanceDataItem) or None
        :return: a summary, or `None` if the report file can't be written
        :rtype: BitFieldSummary
        """
        file_name = os.path.join(report_default_directory, _FILE_NAME)
        if provenance_items is None:
            provenance_items = []
        try:
            with open(file_name, "w") as f:
                return self._write_report(
                    f, provenance_items, machine_graph, placements)
        except IOError:
            logger.exception("Generate_placement_reports: Can't open file"
                             " {} for writing.", _FILE_NAME)
            return None

    @staticmethod
    def _merged_component(provenance_items, to_merge_per_chip, writer):
        """ Report how many bitfields were merged into the router.

        :param list(ProvenanceDataItem) provenance_items: prov items
        :param dict([int, int], int: to_merge_per_chip: number of bitfields
            that could be merged per chip
        :param ~io.FileIO writer: file writer.
        :return: tuple containing 4 elements.
         1. min_bit_fields merged in a chip,
         2. the max bit_fields merged in a chip,
         3. the total bit_fields_merged into all the routers.
         4. average number of bit-fields merged on the routers.
        :rtype: tuple(int or str, int or str, int or str, float or str)
        """
        top_bit_field = 0
        min_bit_field = sys.maxsize
        total_bit_fields_merged = 0
        average_per_chip_merged = 0
        n_chips = 0
        to_merge_chips = set(to_merge_per_chip.keys())

        found = False
        for prov_item in provenance_items:
            if prov_item.names[0] == PROV_TOP_NAME:
                bits = prov_item.names[1].split("_")
                x = int(bits[3])
                y = int(bits[4])
                if (x, y) not in to_merge_per_chip:
                    continue
                to_merge = to_merge_per_chip[x, y]
                merged = int(prov_item.value)
                found = True
                writer.write(
                    "Chip {}:{} has {} bitfields out of {} merged into it."
                    " Which is {:.2%}\n".format(
                        x, y, merged, to_merge, merged / to_merge))
                total_bit_fields_merged += int(prov_item.value)
                if merged > top_bit_field:
                    top_bit_field = merged
                if merged < min_bit_field:
                    min_bit_field = merged
                average_per_chip_merged += merged
                n_chips += 1

        if found:
            average_per_chip_merged = (
                float(average_per_chip_merged) / float(n_chips))
        else:
            min_bit_field = NOT_APPLICABLE
            top_bit_field = NOT_APPLICABLE
            total_bit_fields_merged = NOT_APPLICABLE
            average_per_chip_merged = NOT_APPLICABLE

        if len(to_merge_chips) > 0:
            writer.write(
                "The Chips {} had bitfields. \n"
                "But no record was found of any attepmt to merge them \n"
                "".format(to_merge_chips))

        return (min_bit_field, top_bit_field, total_bit_fields_merged,
                average_per_chip_merged)

    @staticmethod
    def _compute_to_merge_per_chip(machine_graph, placements):
        """
        :param ~.MachineGraph machine_graph:
        :param ~.Placements placements:
        :rtype: tuple(int, int, int, float or int)
        """
        total_to_merge = 0
        to_merge_per_chip = defaultdict(int)

        for placement in placements:
            binary_start_type = None
            if isinstance(placement.vertex, AbstractHasAssociatedBinary):
                binary_start_type = placement.vertex.get_binary_start_type()

            if binary_start_type != ExecutableType.SYSTEM:
                seen_partitions = set()
                for incoming_partition in machine_graph.\
                        get_multicast_edge_partitions_ending_at_vertex(
                            placement.vertex):
                    if incoming_partition not in seen_partitions:
                        total_to_merge += 1
                        to_merge_per_chip[placement.x, placement.y] += 1
                        seen_partitions.add(incoming_partition)

        return total_to_merge, to_merge_per_chip

    @staticmethod
    def _before_merge_component(total_to_merge, to_merge_per_chip):
        """
        :rtype: tuple(int, int, int, float or int)
        """
        max_bit_fields_on_chip = 0
        min_bit_fields_on_chip = sys.maxsize

        for bitfield_count in to_merge_per_chip.values():
            if bitfield_count > max_bit_fields_on_chip:
                max_bit_fields_on_chip = bitfield_count
            if bitfield_count < min_bit_fields_on_chip:
                min_bit_fields_on_chip = bitfield_count

        if len(to_merge_per_chip) == 0:
            average = 0
        else:
            average = float(total_to_merge) / float(len(to_merge_per_chip))

        return max_bit_fields_on_chip, min_bit_fields_on_chip, average

    def _write_report(
            self, writer, provenance_items, machine_graph, placements):
        """ writes the report

        :param ~io.FileIO writer: the file writer
        :param list(ProvenanceDataItem) provenance_items: the prov items
        :param ~.MachineGraph machine_graph: the machine graph
        :param ~.Placements placements: the placements
        :return: a summary
        :rtype: BitFieldSummary
        """
        total_to_merge, to_merge_per_chip = self._compute_to_merge_per_chip(
            machine_graph, placements)
        (max_to_merge_per_chip, low_to_merge_per_chip,
         average_per_chip_to_merge) = self._before_merge_component(
            total_to_merge, to_merge_per_chip)
        (min_bit_field, top_bit_field, total_bit_fields_merged,
         average_per_chip_merged) = self._merged_component(
            provenance_items, to_merge_per_chip, writer)
        writer.write(
            "\n\nBefore merge there where {} bitfields on {} Chips "
            "ranging from {} to {} bitfields per chip with an average of {}"
            "".format(
                total_to_merge, len(to_merge_per_chip), max_to_merge_per_chip,
                low_to_merge_per_chip, average_per_chip_to_merge))
        writer.write(
            "\nSuccessfully merged {} bitfields ranging from {} to {} "
            "bitfields per chip with an average of {}".format(
                total_bit_fields_merged, top_bit_field, min_bit_field,
                average_per_chip_merged))
        if total_to_merge:
            if total_bit_fields_merged == NOT_APPLICABLE:
                writer.write(
                    "\nNone of the {} bitfields merged".format(
                        total_to_merge))
            else:
                writer.write(
                    "\nIn total {:.2%} of the bitfields merged".format(
                        total_bit_fields_merged / total_to_merge))

        return BitFieldSummary(
            lowest_per_chip=min_bit_field, max_per_chip=top_bit_field,
            total_merged=total_bit_fields_merged,
            total_to_merge=total_to_merge,
            max_to_merge_per_chip=max_to_merge_per_chip,
            low_to_merge_per_chip=low_to_merge_per_chip,
            average_per_chip_merged=average_per_chip_merged,
            average_per_chip_to_merge=average_per_chip_to_merge)



# Copyright (c) 2019-2020 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinnman.exceptions import SpinnmanException
from spinnman.messages.scp.enums import Signal
from spinnman.model import ExecutableTargets
from spinn_front_end_common.utilities.utility_objs import ExecutableType
from spinn_front_end_common.utilities import IOBufExtractor


def run_system_application(
        executable_cores, app_id, transceiver, provenance_file_path,
        executable_finder, read_algorithm_iobuf, check_for_success_function,
        cpu_end_states, needs_sync_barrier, filename_template,
        binaries_to_track=None, progress_bar=None, logger=None):
    """ Executes the given _system_ application. \
        Used for on-chip expanders, compressors, etc.

    :param ~spinnman.model.ExecutableTargets executable_cores:
        the cores to run the executable on
    :param int app_id: the app-id for the executable
    :param ~spinnman.transceiver.Transceiver transceiver: the SpiNNMan instance
    :param str provenance_file_path:
        the path for where provenance data is stored
    :param ExecutableFinder executable_finder: finder for executable paths
    :param bool read_algorithm_iobuf: whether to report IOBUFs
    :param callable check_for_success_function:
        function used to check success;
        expects `executable_cores`, `transceiver` as inputs
    :param set(~spinnman.model.enums.CPUState) cpu_end_states:
        the states that a successful run is expected to terminate in
    :param bool needs_sync_barrier: whether a sync barrier is needed
    :param str filename_template: the IOBUF filename template.
    :param list(str) binaries_to_track:
        A list of binary names to check for exit state.
        Or `None` for all binaries
    :param progress_bar: Possible progress bar to update.
           end() will be called after state checked
    :param ~logging.Logger logger:
        If provided and IOBUF is extracted, will be used to log errors and
        warnings
    :type progress_bar: ~spinn_utilities.progress_bar.ProgressBar or None
    :raise SpinnmanException:
        If one should arise from the underlying SpiNNMan calls
    """

    # load the executable
    transceiver.execute_application(executable_cores, app_id)

    if needs_sync_barrier:

        # fire all signals as required
        transceiver.send_signal(app_id, Signal.SYNC0)

    error = None
    binary_start_types = dict()
    if binaries_to_track is None:
        check_targets = executable_cores
    else:
        check_targets = ExecutableTargets()
        for binary_name in binaries_to_track:
            check_targets.add_subsets(
                binary_name,
                executable_cores.get_cores_for_binary(binary_name))
    for binary_name in executable_cores.binaries:
        binary_start_types[binary_name] = ExecutableType.SYSTEM

    # Wait for the executable to finish
    succeeded = False
    try:
        transceiver.wait_for_cores_to_be_in_state(
            check_targets.all_core_subsets, app_id, cpu_end_states,
            progress_bar=progress_bar)
        if progress_bar is not None:
            progress_bar.end()
        succeeded = True
    except SpinnmanException as ex:
        # Delay the exception until iobuf is ready
        error = ex

    if progress_bar is not None:
        progress_bar.end()

    # Check if any cores have not completed successfully
    if succeeded and check_for_success_function:
        succeeded = check_for_success_function(executable_cores, transceiver)

    # if doing iobuf or on failure (succeeded is None is not failure)
    if read_algorithm_iobuf or not succeeded:
        _report_iobuf_messages(
            transceiver, executable_cores, executable_finder, logger,
            filename_template, provenance_file_path)

    # stop anything that's associated with the compressor binary
    transceiver.stop_application(app_id)
    transceiver.app_id_tracker.free_id(app_id)

    if error is not None:
        raise error  # pylint: disable=raising-bad-type


def _report_iobuf_messages(
        txrx, cores, exe_finder, logger, filename_template, directory):
    """
    :param Transceiver txrx:
    :param ~spinnman.model.ExecutableTargets cores:
    :param ExecutableFinder exe_finder:
    :param ~logging.Logger logger:
    :param str filename_template:
    :param str directory:
    """
    # Import in this function to prevent circular import issue
    iobuf_reader = IOBufExtractor(
        txrx, cores, exe_finder, app_provenance_file_path=directory,
        system_provenance_file_path=directory,
        filename_template=filename_template, suppress_progress=False)
    error_entries, warn_entries = iobuf_reader.extract_iobuf()
    if logger is not None:
        for entry in warn_entries:
            logger.warn(entry)
        for entry in error_entries:
            logger.error(entry)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.progress_bar import ProgressBar
from spinnman.constants import MAX_TAG_ID


class TagsLoader(object):
    """ Loads tags onto the machine.
    """

    __slots__ = []

    def __call__(
            self, transceiver, tags=None, iptags=None, reverse_iptags=None):
        """
        :param ~spinnman.transceiver.Transceiver transceiver:
            the transceiver object
        :param ~pacman.model.tags.Tags tags:
            the tags object which contains IP and reverse IP tags;
            could be ``None`` if these are being given in separate lists
        :param list(~spinn_machine.tags.IPTag) iptags:
            a list of IP tags, given when tags is ``None``
        :param list(~spinn_machine.tags.ReverseIPTag) reverse_iptags:
            a list of reverse IP tags when tags is ``None``
        """
        # clear all the tags from the Ethernet connection, as nothing should
        # be allowed to use it (no two apps should use the same Ethernet
        # connection at the same time)
        progress = ProgressBar(MAX_TAG_ID, "Clearing tags")
        for tag_id in progress.over(range(MAX_TAG_ID)):
            transceiver.clear_ip_tag(tag_id)

        # Use tags object to supply tag info if it is supplied
        if tags is not None:
            iptags = list(tags.ip_tags)
            reverse_iptags = list(tags.reverse_ip_tags)

        # Load the IP tags and the Reverse IP tags
        progress = ProgressBar(
            len(iptags) + len(reverse_iptags), "Loading Tags")
        self.load_iptags(iptags, transceiver, progress)
        self.load_reverse_iptags(reverse_iptags, transceiver, progress)
        progress.end()

    @staticmethod
    def load_iptags(iptags, transceiver, progress_bar):
        """ Loads all the IP tags individually.

        :param list(~spinn_machine.tags.IPTag) iptags:
            the IP tags to be loaded.
        :param ~spinnman.transceiver.Transceiver transceiver:
            the transceiver object
        :param ~spinn_utilities.progress_bar.ProgressBar progress_bar:
        """
        for ip_tag in progress_bar.over(iptags, False):
            transceiver.set_ip_tag(ip_tag)

    @staticmethod
    def load_reverse_iptags(reverse_ip_tags, transceiver, progress_bar):
        """ Loads all the reverse IP tags individually.

        :param list(~spinn_machine.tags.ReverseIPTag) reverse_ip_tags:
            the reverse IP tags to be loaded
        :param ~spinnman.transceiver.Transceiver transceiver:
            the transceiver object
        :param ~spinn_utilities.progress_bar.ProgressBar progress_bar:
        """
        for reverse_ip_tag in progress_bar.over(reverse_ip_tags, False):
            transceiver.set_reverse_ip_tag(reverse_ip_tag)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.overrides import overrides
from spinnman.messages.scp import SCPRequestHeader
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
from spinnman.messages.sdp import SDPFlag, SDPHeader
from spinnman.messages.scp.impl.check_ok_response import CheckOKResponse
from spinn_front_end_common.utilities.constants import SDP_PORTS
from .speedup_in_scp_commands import SpeedupInSCPCommands


class LoadSystemMCRoutesMessage(AbstractSCPRequest):
    """ An SCP Request to write the system multicast routes into the router.
    """

    __slots__ = []

    def __init__(self, x, y, p):
        """
        :param int x: The x-coordinate of a chip, between 0 and 255
        :param int y: The y-coordinate of a chip, between 0 and 255
        :param int p:
            The processor running the extra monitor vertex, between 0 and 17
        """

        super().__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED,
                destination_port=(
                    SDP_PORTS.EXTRA_MONITOR_CORE_DATA_IN_SPEED_UP.value),
                destination_cpu=p, destination_chip_x=x,
                destination_chip_y=y),
            SCPRequestHeader(
                command=SpeedupInSCPCommands.LOAD_SYSTEM_MC_ROUTES))

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self):
        return CheckOKResponse(
            "load system multicast routes",
            SpeedupInSCPCommands.LOAD_SYSTEM_MC_ROUTES)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.progress_bar import ProgressBar
from spinn_front_end_common.interface.buffer_management import BufferManager
from spinn_front_end_common.interface.buffer_management.buffer_models \
    import (
        AbstractSendsBuffersFromHost, AbstractReceiveBuffersToHost)


class BufferManagerCreator(object):
    """ Creates a buffer manager.
    """
    __slots__ = []

    def __call__(
            self, placements, tags, txrx,
            uses_advanced_monitors, report_folder, extra_monitor_cores=None,
            extra_monitor_to_chip_mapping=None,
            packet_gather_cores_to_ethernet_connection_map=None, machine=None,
            fixed_routes=None, java_caller=None):
        """
        :param ~pacman.model.placements.Placements placements:
        :param ~pacman.model.tags.Tags tags:
        :param ~spinnman.transceiver.Transceiver txrx:
        :param bool uses_advanced_monitors:
        :param str report_folder:
            The path where the SQLite database holding the data will be placed,
            and where any java provenance can be written.
        :param list(ExtraMonitorSupportMachineVertex) extra_monitor_cores:
        :param extra_monitor_to_chip_mapping:
        :type extra_monitor_to_chip_mapping:
            dict(tuple(int,int),ExtraMonitorSupportMachineVertex)
        :param packet_gather_cores_to_ethernet_connection_map:
        :type packet_gather_cores_to_ethernet_connection_map:
            dict(tuple(int,int),DataSpeedUpPacketGatherMachineVertex)
        :param ~spinn_machine.Machine machine:
        :param fixed_routes:
        :type fixed_routes: dict(tuple(int,int),~spinn_machine.FixedRouteEntry)
        :param JavaCaller java_caller:
        :rtype: BufferManager
        """
        # pylint: disable=too-many-arguments
        progress = ProgressBar(placements.n_placements, "Initialising buffers")

        # Create the buffer manager
        buffer_manager = BufferManager(
            placements=placements, tags=tags, transceiver=txrx,
            extra_monitor_cores=extra_monitor_cores,
            packet_gather_cores_to_ethernet_connection_map=(
                packet_gather_cores_to_ethernet_connection_map),
            extra_monitor_to_chip_mapping=extra_monitor_to_chip_mapping,
            machine=machine, uses_advanced_monitors=uses_advanced_monitors,
            fixed_routes=fixed_routes, report_folder=report_folder,
            java_caller=java_caller)

        for placement in progress.over(placements.placements):
            vertex = placement.vertex
            if isinstance(vertex, AbstractSendsBuffersFromHost) and \
                    vertex.buffering_input():
                buffer_manager.add_sender_vertex(vertex)

            if isinstance(vertex, AbstractReceiveBuffersToHost):
                buffer_manager.add_receiving_vertex(vertex)

        return buffer_manager



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from spinn_utilities.progress_bar import ProgressBar
from pacman.model.graphs.application import ApplicationEdge
from pacman.model.graphs.machine import MachineEdge
from spinn_front_end_common.utilities.constants import (
    PARTITION_ID_FOR_MULTICAST_DATA_SPEED_UP)
from spinn_front_end_common.utility_models import (
    DataSpeedUpPacketGatherMachineVertex as DataSpeedUp,
    ExtraMonitorSupportMachineVertex)


class InsertEdgesToExtraMonitorFunctionality(object):
    """ Inserts edges between vertices who use MC speed up and its local\
        MC data gatherer.
    """

    __slots__ = [
        # Map of Chip(x,y) to the (Gather) vertex
        "_chip_to_gatherer_map",
        # Description of the machine on which job is being run
        "_machine",
        # Map of Core(x,y,p) to the Placement on that core
        "_placements"
    ]

    EDGE_LABEL = "edge between {} and {}"

    def __call__(self, machine_graph, placements, machine,
                 vertex_to_ethernet_connected_chip_mapping,
                 application_graph=None):
        """
        :param ~pacman.model.graphs.machine.MachineGraph machine_graph:
            the machine graph instance
        :param ~pacman.model.placements.Placements placements: the placements
        :param ~spinn_machine.Machine machine: the machine object
        :param vertex_to_ethernet_connected_chip_mapping:
            mapping between ethernet connected chips and packet gatherers
        :type vertex_to_ethernet_connected_chip_mapping:
            dict(tuple(int,int), DataSpeedUpPacketGatherMachineVertex)
        :param application_graph: the application graph
        :type application_graph:
            ~pacman.model.graphs.application.ApplicationGraph
        """
        # pylint: disable=too-many-arguments, attribute-defined-outside-init
        n_app_vertices = 0
        if application_graph is not None:
            n_app_vertices = application_graph.n_vertices
        self._chip_to_gatherer_map = vertex_to_ethernet_connected_chip_mapping
        self._machine = machine
        self._placements = placements

        progress = ProgressBar(
            machine_graph.n_vertices + n_app_vertices,
            "Inserting edges between vertices which require FR speed up "
            "functionality.")

        if application_graph is None:
            for vertex in progress.over(machine_graph.vertices):
                if isinstance(vertex, ExtraMonitorSupportMachineVertex):
                    self._process_mach_graph_vertex(vertex, machine_graph)
        else:
            for vertex in progress.over(machine_graph.vertices, False):
                if isinstance(vertex, ExtraMonitorSupportMachineVertex):
                    self._process_app_graph_vertex(
                        vertex, machine_graph, application_graph)

    def _process_app_graph_vertex(
            self, vertex, machine_graph, application_graph):
        """ Inserts edges as required for a given vertex

        :param ExtraMonitorSupportMachineVertex vertex: the extra monitor core
        :param ~.MachineGraph machine_graph: machine graph object
        :param ~.ApplicationGraph application_graph: app graph object; not None
        :rtype: None
        """
        gatherer = self._get_gatherer_vertex(vertex)

        # locate if edge is already built; if not, build it and do mapping
        if not self.__has_edge_already(vertex, gatherer, machine_graph):
            # See if there's an application edge already
            app_edge = self.__get_app_edge(
                application_graph, vertex.app_vertex, gatherer.app_vertex)
            if app_edge is None:
                app_edge = ApplicationEdge(
                    vertex.app_vertex, gatherer.app_vertex)
                application_graph.add_edge(
                    app_edge, PARTITION_ID_FOR_MULTICAST_DATA_SPEED_UP)
            # Use the application edge to build the machine edge
            edge = MachineEdge(
                vertex, gatherer, traffic_type=DataSpeedUp.TRAFFIC_TYPE,
                label=self.EDGE_LABEL.format(vertex, gatherer),
                app_edge=app_edge)
            machine_graph.add_edge(
                edge, PARTITION_ID_FOR_MULTICAST_DATA_SPEED_UP)

    def _process_mach_graph_vertex(self, vertex, machine_graph):
        """ Inserts edges as required for a given vertex

        :param ExtraMonitorSupportMachineVertex vertex: the extra monitor core
        :param ~.MachineGraph machine_graph:
            machine graph object, which is not associated with any application
            graph
        :rtype: None
        """
        gatherer = self._get_gatherer_vertex(vertex)

        # locate if edge is already built; if not, build it and do mapping
        if not self.__has_edge_already(vertex, gatherer, machine_graph):
            edge = MachineEdge(
                vertex, gatherer, traffic_type=DataSpeedUp.TRAFFIC_TYPE)
            machine_graph.add_edge(
                edge, PARTITION_ID_FOR_MULTICAST_DATA_SPEED_UP)

    def _get_gatherer_vertex(self, vertex):
        """
        :param ExtraMonitorSupportMachineVertex vertex:
        :rtype: DataSpeedUpPacketGatherMachineVertex
        """
        placement = self._placements.get_placement_of_vertex(vertex)
        chip = self._machine.get_chip_at(placement.x, placement.y)
        ethernet_chip = self._machine.get_chip_at(
            chip.nearest_ethernet_x, chip.nearest_ethernet_y)
        return self._chip_to_gatherer_map[ethernet_chip.x, ethernet_chip.y]

    @staticmethod
    def __get_app_edge(graph, source, destination):
        """
        :param ~.ApplicationGraph graph:
        :param ~.ApplicationVertex source:
        :param ~.ApplicationVertex destination:
        :rtype: ~.ApplicationEdge
        """
        for edge in graph.get_edges_ending_at_vertex(destination):
            if edge.pre_vertex == source:
                return edge
        return None

    @staticmethod
    def __has_edge_already(source, destination, graph):
        """ Checks if a edge already exists

        :param ~.AbstractVertex source: the source of the edge
        :param ~.AbstractVertex destination: destination of the edge
        :param ~.Graph graph: which graph to look in
        :return: Whether the edge was found
        :rtype: bool
        """
        return any(edge.pre_vertex == source
                   for edge in graph.get_edges_ending_at_vertex(destination))



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from spinn_utilities.progress_bar import ProgressBar


class BoardChipReport(object):
    """ Report on memory usage
    """

    AREA_CODE_REPORT_NAME = "board_chip_report.txt"

    def __call__(self, report_default_directory, machine):
        """ Creates a report that states where in SDRAM each region is.

        :param str report_default_directory:
            the folder where reports are written
        :param ~spinn_machine.Machine machine:
            python representation of the machine
        :rtype: None
        """

        # create file path
        directory_name = os.path.join(
            report_default_directory, self.AREA_CODE_REPORT_NAME)

        # create the progress bar for end users
        progress_bar = ProgressBar(
            len(machine.ethernet_connected_chips),
            "Writing the board chip report")

        # iterate over ethernet chips and then the chips on that board
        with open(directory_name, "w") as writer:
            self._write_report(writer, machine, progress_bar)

    @staticmethod
    def _write_report(writer, machine, progress_bar):
        """
        :param ~io.FileIO writer:
        :param ~spinn_machine.Machine machine:
        :param ~spinn_utilities.progress_bar.ProgressBar progress_bar:
        """
        for chip in progress_bar.over(machine.ethernet_connected_chips):
            xys = machine.get_existing_xys_on_board(chip)
            writer.write(
                "board with IP address : {} : has chips {}\n".format(
                    chip.ip_address, list(xys)))



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import io


class DataRowWriter(io.RawIOBase):
    __slots__ = [
        "_x",
        "_y",
        "_p",
        "_targets",
        "_data",
    ]

    def __init__(self, x, y, p, targets):
        super().__init__()
        self._x = x
        self._y = y
        self._p = p
        self._targets = targets
        self._data = bytearray()

    def write(self, data):
        assert self.closed is False
        self._data += data

    def readable(self):
        return False

    def seekable(self):
        return False

    def writable(self):
        return False

    def truncate(self, size=None):
        # Ignore
        pass

    def fileno(self):
        raise OSError

    def close(self):
        """ Closes the writer if not already closed.
        """
        if not self.closed:
            self._targets.write_data_spec(
                self._x, self._y, self._p, self._data)
        super().close()



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

_REPORT_FILENAME = "tags_on_machine.txt"


class TagsFromMachineReport(object):
    """ Describes what the tags actually present on the machine are.
    """

    def __call__(self, report_default_directory, transceiver):
        """
        :param str report_default_directory:
        :param ~spinnman.transceiver.Transceiver transceiver:
        """
        filename = os.path.join(report_default_directory, _REPORT_FILENAME)
        tags = self._get_tags(transceiver)
        with open(filename, "w") as f:
            f.write("Tags actually read off the machine\n")
            f.write("==================================\n")
            for tag in tags:
                f.write("{}\n".format(self._render_tag(tag)))

    @staticmethod
    def _get_tags(txrx):
        try:
            return txrx.get_tags()
        except Exception as e:  # pylint: disable=broad-except
            return [e]

    @staticmethod
    def _render_tag(tag):
        return repr(tag)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pacman.executor import injection_decorator

# pylint: disable=global-statement
_failed_state = None
_simulator = None
_cached_simulator = None


def get_simulator():
    """ Get the current simulator object.

    :rtype: SimulatorInterface
    """
    global _simulator, _failed_state
    if _simulator is None:
        if _failed_state is None:
            raise ValueError("You must import one of the simulator classes "
                             "before calling get_simulator")
        return _failed_state
    return _simulator


def get_not_running_simulator():
    """ Get the current simulator object and verify that it is not running.

    :rtype: SimulatorInterface
    """
    global _simulator, _failed_state
    if _simulator is None:
        if _failed_state is None:
            raise ValueError("You must import one of the simulator classes "
                             "before calling get_simulator")
        return _failed_state
    _simulator.verify_not_running()
    return _simulator


def set_simulator(new_simulator):
    """ Set the current simulator object.

    :param SimulatorInterface new_simulator: The simulator to set.
    """
    global _simulator, _failed_state, _cached_simulator
    if _failed_state is None:
        raise ValueError("Unexpected call to set_simulator before "
                         "set_failed_state")
    _simulator = new_simulator
    _cached_simulator = None


def unset_simulator(to_cache_simulator=None):
    """ Destroy the current simulator.

    :param SimulatorInterface to_cache_simulator:
        a cached version for allowing data extraction
    """
    global _simulator, _cached_simulator
    _simulator = None
    _cached_simulator = to_cache_simulator

    injection_decorator._instances = list()


def has_simulator():
    """ Check if a simulator is operational.

    :rtype: bool
    """
    global _simulator
    return _simulator is not None


def set_failed_state(new_failed_state):
    """ Install a marker to say that the simulator has failed.

    :param FailedState new_failed_state: the failure marker
    """
    # pylint: disable=unidiomatic-typecheck
    global _failed_state
    if _failed_state is None:
        _failed_state = new_failed_state
    elif type(new_failed_state) != type(_failed_state):
        raise ValueError("You may only setup/init one type of simulator")


def _last_simulator():
    """ Get last simulator to be used.

    Before setup this will return None.
    Between setup and end this will return the simulator.
    After end this will return the previous simulator
    """
    global _simulator, _cached_simulator
    if _simulator is not None:
        return _simulator
    if _cached_simulator is not None:
        return _cached_simulator
    if _failed_state is not None:
        return None
    else:
        raise ValueError(
            "There should be some sort of simulator set. Why am i here?!")


def get_generated_output(output):
    """ Get one of the simulator outputs by name.

    :param str output: The name of the output to retrieve.
    :return: The value (of arbitrary type, dependent on which output),
        or `None` if the variable is not found.
    :raises ValueError:
        if the system is in a state where outputs can't be retrieved
    """
    simulator = _last_simulator()
    if simulator is None:
        raise ValueError(
            "You need to have ran a simulator before asking for its "
            "generated output.")
    else:
        return simulator.get_generated_output(output)


def provenance_file_path():
    """
    Returns the path to the directory that holds all provenance files

    This will be the path used by the last run call or to be used by
    the next run if it has not yet been called.

    :rtpye: str
    :raises ValueError:
        if the system is in a state where path can't be retrieved
    """
    simulator = _last_simulator()
    if simulator is None:
        raise ValueError(
            "You need to have setup a simulator before asking for its "
            "provenance_file_path.")
    else:
        # underscore param used avoid exposing a None PyNN parameter
        return simulator._provenance_file_path


def app_provenance_file_path():
    """
    Returns the path to the directory that holds all app provenance files

    This will be the path used by the last run call or to be used by
    the next run if it has not yet been called.

    :rtpye: str
    :raises ValueError:
        if the system is in a state where path can't be retrieved
    """
    simulator = _last_simulator()
    if simulator is None:
        raise ValueError(
            "You need to have setup a simulator before asking for its "
            "app_provenance_file_path.")
    else:
        # underscore param used avoid exposing a None PyNN parameter
        return simulator._app_provenance_file_path


def system_provenance_file_path():
    """
    Returns the path to the directory that holds all provenance files

    This will be the path used by the last run call or to be used by
    the next run if it has not yet been called.

    :rtpye: str
    :raises ValueError:
        if the system is in a state where path can't be retrieved
    """
    simulator = _last_simulator()
    if simulator is None:
        raise ValueError(
            "You need to have setup a simulator before asking for its "
            "system_provenance_file_path.")
    else:
        # underscore param used avoid exposing a None PyNN parameter
        return simulator._system_provenance_file_path


def run_report_directory():
    """
    Returns the path to the directory that holds all the reports for run

    This will be the path used by the last run call or to be used by
    the next run if it has not yet been called.

    :rtpye: str
    :raises ValueError:
        if the system is in a state where path can't be retrieved
    """
    simulator = _last_simulator()
    if simulator is None:
        raise ValueError(
            "You need to have setup a simulator before asking for its "
            "run_report_directory.")
    else:
        # underscore param used avoid exposing a None PyNN parameter
        return simulator._report_default_directory


def config():
    """ Provides access to the configuration used or a best effort to the
    config to be used

    If called before setup will return the values found in the config file
    BEFORE any cleanup alterations that setup could do.

    After setup will return the simulation config even if end has been called

    :rtype: ~spinn_front_end_common.interface.config_handler.ConfigHandler
    """
    global _failed_state
    simulator = _last_simulator()
    if simulator is None:
        return _failed_state.config
    else:
        return simulator.config



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_front_end_common.utilities.scp import ClearIOBUFProcess
from spinn_front_end_common.utilities.utility_objs import ExecutableType


class ChipIOBufClearer(object):
    """ Clears the logging output buffer of an application running on a\
        SpiNNaker machine.
    """

    __slots__ = []

    def __call__(self, transceiver, executable_types):
        """
        :param ~spinnman.transceiver.Transceiver transceiver:
        :param executable_types:
        :type executable_types:
            dict(ExecutableType,~spinn_machine.CoreSubsets)
        """
        if ExecutableType.USES_SIMULATION_INTERFACE in executable_types:
            core_subsets = \
                executable_types[ExecutableType.USES_SIMULATION_INTERFACE]

            process = ClearIOBUFProcess(transceiver.scamp_connection_selector)
            process.clear_iobuf(core_subsets, len(core_subsets))



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import threading
import ctypes
from spinn_utilities.log import FormatAdapter
from spinn_utilities.ordered_set import OrderedSet
from spinn_utilities.progress_bar import ProgressBar
from spinn_utilities.timer import Timer
from spinnman.constants import UDP_MESSAGE_MAX_SIZE
from spinnman.connections.udp_packet_connections import EIEIOConnection
from spinnman.messages.eieio.command_messages import (
    EIEIOCommandMessage, StopRequests, SpinnakerRequestBuffers,
    HostSendSequencedData, EventStopRequest)
from spinnman.utilities import utility_functions
from spinnman.messages.sdp import SDPHeader, SDPMessage, SDPFlag
from spinnman.messages.eieio import EIEIOType
from spinnman.messages.eieio.data_messages import EIEIODataMessage
from data_specification.constants import BYTES_PER_WORD
from spinn_front_end_common.utilities.constants import SDP_PORTS
from spinn_front_end_common.utilities.exceptions import (
    BufferableRegionTooSmall, SpinnFrontEndException)
from spinn_front_end_common.utilities.helpful_functions import (
    locate_memory_region_for_placement, locate_extra_monitor_mc_receiver)
from spinn_front_end_common.utilities.globals_variables import get_simulator
from spinn_front_end_common.interface.buffer_management.storage_objects \
    import (BuffersSentDeque, BufferedReceivingData)
from spinn_front_end_common.interface.buffer_management.buffer_models \
    import AbstractReceiveBuffersToHost
from .recording_utilities import get_recording_header_size

logger = FormatAdapter(logging.getLogger(__name__))

# The minimum size of any message - this is the headers plus one entry
_MIN_MESSAGE_SIZE = EIEIODataMessage.min_packet_length(
    eieio_type=EIEIOType.KEY_32_BIT, is_timestamp=True)

# The number of bytes in each key to be sent
_N_BYTES_PER_KEY = EIEIOType.KEY_32_BIT.key_bytes  # @UndefinedVariable

_SDP_MAX_PACKAGE_SIZE = 272

TRAFFIC_IDENTIFIER = "BufferTraffic"

VERIFY = False


class _RecordingRegion(ctypes.LittleEndianStructure):
    """ Recording Region data
    """
    _fields_ = [
        # Space available for recording
        ("space", ctypes.c_uint32),
        # The size of the recording region
        ("size", ctypes.c_uint32, 31),
        # Whether any data is missing
        ("missing", ctypes.c_uint32, 1),
        # The address of the data
        ("data", ctypes.c_uint32)
    ]


class BufferManager(object):
    """ Manager of send buffers.
    """

    __slots__ = [
        # placements object
        "_placements",

        # list of tags
        "_tags",

        # SpiNNMan instance
        "_transceiver",

        # Set of (ip_address, port) that are being listened to for the tags
        "_seen_tags",

        # Set of vertices with buffers to be sent
        "_sender_vertices",

        # Dictionary of sender vertex -> buffers sent
        "_sent_messages",

        # storage area for received data from cores
        "_received_data",

        # Lock to avoid multiple messages being processed at the same time
        "_thread_lock_buffer_out",

        # Lock to avoid multiple messages being processed at the same time
        "_thread_lock_buffer_in",

        # bool flag
        "_finished",

        # listener port
        "_listener_port",

        # the extra monitor cores which support faster data extraction
        "_extra_monitor_cores",

        # the extra_monitor to Ethernet connection map
        "_packet_gather_cores_to_ethernet_connection_map",

        # monitor cores via chip ID
        "_extra_monitor_cores_by_chip",

        # fixed routes, used by the speed up functionality for reports
        "_fixed_routes",

        # machine object
        "_machine",

        # flag for what data extraction to use
        "_uses_advanced_monitors",

        # Support class to help call Java
        "_java_caller"
    ]

    def __init__(self, placements, tags, transceiver, extra_monitor_cores,
                 packet_gather_cores_to_ethernet_connection_map,
                 extra_monitor_to_chip_mapping, machine, fixed_routes,
                 uses_advanced_monitors, report_folder, java_caller=None):
        """
        :param ~pacman.model.placements.Placements placements:
            The placements of the vertices
        :param ~pacman.model.tags.Tags tags: The tags assigned to the vertices
        :param ~spinnman.transceiver.Transceiver transceiver:
            The transceiver to use for sending and receiving information
        :param list(ExtraMonitorSupportMachineVertex) extra_monitor_cores:
            The monitors.
        :param packet_gather_cores_to_ethernet_connection_map:
            mapping of cores to the gatherer vertex placed on them
        :type packet_gather_cores_to_ethernet_connection_map:
            dict(tuple(int,int), DataSpeedUpPacketGatherMachineVertex)
        :param extra_monitor_to_chip_mapping:
        :type extra_monitor_to_chip_mapping:
            dict(tuple(int,int),ExtraMonitorSupportMachineVertex)
        :param ~spinn_machine.Machine machine:
        :param fixed_routes:
        :type fixed_routes: dict(tuple(int,int),~spinn_machine.FixedRouteEntry)
        :param bool uses_advanced_monitors:
        :param str report_folder:
            The directory for reports which includes the file to use as an SQL
            database.
        :param JavaCaller java_caller:
            Support class to call Java, or ``None`` to use Python
        """
        # pylint: disable=too-many-arguments
        self._placements = placements
        self._tags = tags
        self._transceiver = transceiver
        self._extra_monitor_cores = extra_monitor_cores
        self._packet_gather_cores_to_ethernet_connection_map = \
            packet_gather_cores_to_ethernet_connection_map
        self._extra_monitor_cores_by_chip = extra_monitor_to_chip_mapping
        self._fixed_routes = fixed_routes
        self._machine = machine
        self._uses_advanced_monitors = uses_advanced_monitors

        # Set of (ip_address, port) that are being listened to for the tags
        self._seen_tags = set()

        # Set of vertices with buffers to be sent
        self._sender_vertices = set()

        # Dictionary of sender vertex -> buffers sent
        self._sent_messages = dict()

        # storage area for received data from cores
        self._received_data = BufferedReceivingData(report_folder)

        # Lock to avoid multiple messages being processed at the same time
        self._thread_lock_buffer_out = threading.RLock()
        self._thread_lock_buffer_in = threading.RLock()

        self._finished = False
        self._listener_port = None
        self._java_caller = java_caller
        if self._java_caller is not None:
            self._java_caller.set_machine(machine)
            self._java_caller.set_report_folder(report_folder)
            if self._uses_advanced_monitors:
                self._java_caller.set_advanced_monitors(
                    self._placements, self._tags,
                    self._extra_monitor_cores_by_chip,
                    self._packet_gather_cores_to_ethernet_connection_map)

    def _request_data(self, placement_x, placement_y, address, length):
        """ Uses the extra monitor cores for data extraction.

        :param int placement_x:
            the placement x coord where data is to be extracted from
        :param int placement_y:
            the placement y coord where data is to be extracted from
        :param int address: the memory address to start at
        :param int length: the number of bytes to extract
        :return: data as a byte array
        :rtype: bytearray
        """
        # pylint: disable=too-many-arguments
        if not self._uses_advanced_monitors:
            return self._transceiver.read_memory(
                placement_x, placement_y, address, length)

        # Round to word boundaries
        initial = address % BYTES_PER_WORD
        address -= initial
        length += initial
        final = (BYTES_PER_WORD - (length % BYTES_PER_WORD)) % BYTES_PER_WORD
        length += final

        sender = self._extra_monitor_cores_by_chip[placement_x, placement_y]
        receiver = locate_extra_monitor_mc_receiver(
            self._machine, placement_x, placement_y,
            self._packet_gather_cores_to_ethernet_connection_map)
        extra_mon_data = receiver.get_data(
            sender, self._placements.get_placement_of_vertex(sender),
            address, length, self._fixed_routes)
        if VERIFY:
            txrx_data = self._transceiver.read_memory(
                placement_x, placement_y, address, length)
            self._verify_data(extra_mon_data, txrx_data)

        # If we rounded to word boundaries, strip the padding junk
        if initial and final:
            return extra_mon_data[initial:-final]
        elif initial:
            return extra_mon_data[initial:]
        elif final:
            return extra_mon_data[:-final]
        else:
            return extra_mon_data

    def _verify_data(self, extra_mon_data, txrx_data):
        for index, (extra_mon_element, txrx_element) in enumerate(
                zip(extra_mon_data, txrx_data)):
            if extra_mon_element != txrx_element:
                raise Exception("WRONG (at index {})".format(index))

    def _receive_buffer_command_message(self, packet):
        """ Handle an EIEIO command message for the buffers.

        :param packet: The EIEIO message received
        :type packet:
            ~spinnman.messages.eieio.command_messages.EIEIOCommandMessage
        """
        # pylint: disable=broad-except
        if isinstance(packet, SpinnakerRequestBuffers):
            # noinspection PyBroadException
            try:
                self.__request_buffers(packet)
            except Exception:
                logger.exception("problem when sending messages")
        elif isinstance(packet, EIEIOCommandMessage):
            logger.error(
                "The command packet is invalid for buffer management: "
                "command ID {}", packet.eieio_header.command)
        else:
            logger.error(
                "The command packet is invalid for buffer management")

    # Factored out of receive_buffer_command_message to keep code readable
    def __request_buffers(self, packet):
        """
        :param packet: The EIEIO message received
        :type packet:
            ~spinnman.messages.eieio.command_messages.SpinnakerRequestBuffers
        """
        if not self._finished:
            with self._thread_lock_buffer_in:
                vertex = self._placements.get_vertex_on_processor(
                    packet.x, packet.y, packet.p)
                if vertex in self._sender_vertices:
                    self._send_messages(
                        packet.space_available, vertex,
                        packet.region_id, packet.sequence_no)

    def _create_connection(self, tag):
        """
        :param ~spinn_machine.tags.IPTag tag:
        :rtype: ~spinnman.connections.udp_packet_connections.EIEIOConnection
        """
        connection = self._transceiver.register_udp_listener(
            self._receive_buffer_command_message, EIEIOConnection,
            local_port=tag.port, local_host=tag.ip_address)
        self._seen_tags.add((tag.ip_address, connection.local_port))
        utility_functions.send_port_trigger_message(
            connection, tag.board_address)
        logger.info(
            "Listening for packets using tag {} on {}:{}",
            tag.tag, connection.local_ip_address, connection.local_port)
        return connection

    def _add_buffer_listeners(self, vertex):
        """ Add listeners for buffered data for the given vertex

        :param ~pacman.model.graphs.machine.MachineVertex vertex:
        """

        # Find a tag for receiving buffer data
        tags = self._tags.get_ip_tags_for_vertex(vertex)

        if tags is not None:
            # locate tag associated with the buffer manager traffic
            for tag in tags:
                if tag.traffic_identifier == TRAFFIC_IDENTIFIER:
                    # If the tag port is not assigned create a connection and
                    # assign the port.  Note that this *should* update the
                    # port number in any tags being shared.
                    if tag.port is None:
                        # If connection already setup, ensure subsequent
                        # boards use same listener port in their tag
                        if self._listener_port is None:
                            connection = self._create_connection(tag)
                            tag.port = connection.local_port
                            self._listener_port = connection.local_port
                        else:
                            tag.port = self._listener_port

                    # In case we have tags with different specified ports,
                    # also allow the tag to be created here
                    elif (tag.ip_address, tag.port) not in self._seen_tags:
                        self._create_connection(tag)

    def add_receiving_vertex(self, vertex):
        """ Add a vertex into the managed list for vertices which require\
            buffers to be received from them during runtime.

        :param AbstractReceiveBuffersToHost vertex: the vertex to be managed
        """
        self._add_buffer_listeners(vertex)

    def add_sender_vertex(self, vertex):
        """ Add a vertex into the managed list for vertices which require\
            buffers to be sent to them during runtime.

        :param AbstractSendsBuffersFromHost vertex: the vertex to be managed
        """
        self._sender_vertices.add(vertex)
        self._add_buffer_listeners(vertex)

    def load_initial_buffers(self):
        """ Load the initial buffers for the senders using memory writes.
        """
        total_data = 0
        for vertex in self._sender_vertices:
            for region in vertex.get_regions():
                total_data += vertex.get_region_buffer_size(region)

        progress = ProgressBar(total_data, "Loading buffers")
        for vertex in self._sender_vertices:
            for region in vertex.get_regions():
                self._send_initial_messages(vertex, region, progress)
        progress.end()

    def reset(self):
        """ Resets the buffered regions to start transmitting from the\
            beginning of its expected regions and clears the buffered out\
            data files.
        """
        self._received_data.reset()

        # rewind buffered in
        for vertex in self._sender_vertices:
            for region in vertex.get_regions():
                vertex.rewind(region)

        self._finished = False

    def resume(self):
        """ Resets any data structures needed before starting running again.
        """

        # update the received data items
        self._received_data.resume()
        self._finished = False

    def clear_recorded_data(self, x, y, p, recording_region_id):
        """ Removes the recorded data stored in memory.

        :param int x: placement x coordinate
        :param int y: placement y coordinate
        :param int p: placement p coordinate
        :param int recording_region_id: the recording region ID
        """
        self._received_data.clear(x, y, p, recording_region_id)

    def _create_message_to_send(self, size, vertex, region):
        """ Creates a single message to send with the given boundaries.

        :param int size: The number of bytes available for the whole packet
        :param AbstractSendsBuffersFromHost vertex:
            The vertex to get the keys from
        :param int region: The region of the vertex to get keys from
        :return: A new message, or None if no keys can be added
        :rtype: None or ~spinnman.messages.eieio.data_messages.EIEIODataMessage
        """

        # If there are no more messages to send, return None
        if not vertex.is_next_timestamp(region):
            return None

        # Create a new message
        next_timestamp = vertex.get_next_timestamp(region)
        message = EIEIODataMessage.create(
            EIEIOType.KEY_32_BIT, timestamp=next_timestamp)

        # If there is no room for the message, return None
        if message.size + _N_BYTES_PER_KEY > size:
            return None

        # Add keys up to the limit
        bytes_to_go = size - message.size
        while (bytes_to_go >= _N_BYTES_PER_KEY and
                vertex.is_next_key(region, next_timestamp)):

            key = vertex.get_next_key(region)
            message.add_key(key)
            bytes_to_go -= _N_BYTES_PER_KEY

        return message

    def _send_initial_messages(self, vertex, region, progress):
        """ Send the initial set of messages.

        :param AbstractSendsBuffersFromHost vertex:
            The vertex to get the keys from
        :param int region: The region to get the keys from
        :return: A list of messages
        :rtype: list(~spinnman.messages.eieio.data_messages.EIEIODataMessage)
        """

        # Get the vertex load details
        # region_base_address = self._locate_region_address(region, vertex)
        placement = self._placements.get_placement_of_vertex(vertex)
        region_base_address = locate_memory_region_for_placement(
            placement, region, self._transceiver)

        # Add packets until out of space
        sent_message = False
        bytes_to_go = vertex.get_region_buffer_size(region)
        if bytes_to_go % 2 != 0:
            raise SpinnFrontEndException(
                "The buffer region of {} must be divisible by 2".format(
                    vertex))
        all_data = b""
        if vertex.is_empty(region):
            sent_message = True
        else:
            min_size_of_packet = _MIN_MESSAGE_SIZE
            while (vertex.is_next_timestamp(region) and
                    bytes_to_go > min_size_of_packet):
                space_available = min(bytes_to_go, _SDP_MAX_PACKAGE_SIZE)
                next_message = self._create_message_to_send(
                    space_available, vertex, region)
                if next_message is None:
                    break

                # Write the message to the memory
                data = next_message.bytestring
                all_data += data
                sent_message = True

                # Update the positions
                bytes_to_go -= len(data)
                progress.update(len(data))

        if not sent_message:
            raise BufferableRegionTooSmall(
                "The buffer size {} is too small for any data to be added for"
                " region {} of vertex {}".format(bytes_to_go, region, vertex))

        # If there are no more messages and there is space, add a stop request
        if (not vertex.is_next_timestamp(region) and
                bytes_to_go >= EventStopRequest.get_min_packet_length()):
            data = EventStopRequest().bytestring
            # logger.debug(
            #    "Writing stop message of {} bytes to {} on {}, {}, {}".format(
            #         len(data), hex(region_base_address),
            #         placement.x, placement.y, placement.p))
            all_data += data
            bytes_to_go -= len(data)
            progress.update(len(data))
            self._sent_messages[vertex] = BuffersSentDeque(
                region, sent_stop_message=True)

        # Do the writing all at once for efficiency
        self._transceiver.write_memory(
            placement.x, placement.y, region_base_address, all_data)

    def _send_messages(self, size, vertex, region, sequence_no):
        """ Send a set of messages.

        :param int size:
        :param AbstractSendsBuffersFromHost vertex:
        :param int region:
        :param int sequence_no:
        """
        # Get the sent messages for the vertex
        if vertex not in self._sent_messages:
            self._sent_messages[vertex] = BuffersSentDeque(region)
        sent_messages = self._sent_messages[vertex]

        # If the sequence number is outside the window, return now with no
        # messages sent
        if not sent_messages.update_last_received_sequence_number(sequence_no):
            return

        # Remote the existing packets from the size available
        bytes_to_go = size
        for message in sent_messages.messages:
            if isinstance(message.eieio_data_message, EIEIODataMessage):
                bytes_to_go -= message.eieio_data_message.size
            else:
                bytes_to_go -= (message.eieio_data_message
                                .get_min_packet_length())

        # Add messages up to the limits
        while (vertex.is_next_timestamp(region) and
                not sent_messages.is_full and bytes_to_go > 0):

            space_available = min(
                bytes_to_go,
                UDP_MESSAGE_MAX_SIZE -
                HostSendSequencedData.get_min_packet_length())
            # logger.debug(
            #     "Bytes to go {}, space available {}".format(
            #         bytes_to_go, space_available))
            next_message = self._create_message_to_send(
                space_available, vertex, region)
            if next_message is None:
                break
            sent_messages.add_message_to_send(next_message)
            bytes_to_go -= next_message.size
            # logger.debug("Adding additional buffer of {} bytes".format(
            #     next_message.size))

        # If the vertex is empty, send the stop messages if there is space
        if (not sent_messages.is_full and
                not vertex.is_next_timestamp(region) and
                bytes_to_go >= EventStopRequest.get_min_packet_length()):
            sent_messages.send_stop_message()

        # If there are no more messages, turn off requests for more messages
        if not vertex.is_next_timestamp(region) and sent_messages.is_empty():
            # logger.debug("Sending stop")
            self._send_request(vertex, StopRequests())

        # Send the messages
        for message in sent_messages.messages:
            # logger.debug("Sending message with sequence {}".format(
            #     message.sequence_no))
            self._send_request(vertex, message)

    def _send_request(self, vertex, message):
        """ Sends a request.

        :param AbstractSendsBuffersFromHost vertex: The vertex to send to
        :param message: The message to send
        :type message:
            ~spinman.messages.eieio.command_messages.EIEIOCommandMessage
        """

        placement = self._placements.get_placement_of_vertex(vertex)
        sdp_header = SDPHeader(
            destination_chip_x=placement.x, destination_chip_y=placement.y,
            destination_cpu=placement.p, flags=SDPFlag.REPLY_NOT_EXPECTED,
            destination_port=SDP_PORTS.INPUT_BUFFERING_SDP_PORT.value)
        sdp_message = SDPMessage(sdp_header, message.bytestring)
        self._transceiver.send_sdp_message(sdp_message)

    def stop(self):
        """ Indicates that the simulation has finished, so no further\
            outstanding requests need to be processed.
        """
        with self._thread_lock_buffer_in:
            with self._thread_lock_buffer_out:
                self._finished = True

    def get_data_for_placements(self, placements, progress=None):
        """
        :param ~pacman.model.placements.Placements placements:
            Where to get the data from.
        :param progress: How to measure/display the progress.
        :type progress: ~spinn_utilities.progress_bar.ProgressBar or None
        """
        if self._java_caller is not None:
            self._java_caller.set_placements(placements, self._transceiver)

        timer = Timer()
        with timer:
            with self._thread_lock_buffer_out:
                if self._java_caller is not None:
                    self._java_caller.get_all_data()
                    if progress:
                        progress.end()
                elif self._uses_advanced_monitors:
                    self.__old_get_data_for_placements_with_monitors(
                        placements, progress)
                else:
                    self.__old_get_data_for_placements(placements, progress)
        get_simulator().add_extraction_timing(timer.measured_interval)

    def __old_get_data_for_placements_with_monitors(
            self, placements, progress):
        """
        :param ~pacman.model.placements.Placements placements:
            Where to get the data from.
        :param progress: How to measure/display the progress.
        :type progress: ~spinn_utilities.progress_bar.ProgressBar or None
        """
        # locate receivers
        receivers = list(OrderedSet(
            locate_extra_monitor_mc_receiver(
                self._machine, placement.x, placement.y,
                self._packet_gather_cores_to_ethernet_connection_map)
            for placement in placements))

        # update transaction id from the machine for all extra monitors
        for extra_mon in self._extra_monitor_cores:
            extra_mon.update_transaction_id_from_machine(self._transceiver)

        # Ugly, to avoid an import loop...
        with receivers[0].streaming(
                receivers, self._transceiver, self._extra_monitor_cores,
                self._placements):
            # get data
            self.__old_get_data_for_placements(placements, progress)

    def __old_get_data_for_placements(self, placements, progress):
        """
        :param ~pacman.model.placements.Placements placements:
            Where to get the data from.
        :param progress: How to measure/display the progress.
        :type progress: ~spinn_utilities.progress_bar.ProgressBar or None
        """
        # get data
        for placement in placements:
            vertex = placement.vertex
            for recording_region_id in vertex.get_recorded_region_ids():
                self._retreive_by_placement(placement, recording_region_id)
                if progress is not None:
                    progress.update()

    def get_data_by_placement(self, placement, recording_region_id):
        """ Get the data container for all the data retrieved\
            during the simulation from a specific region area of a core.

        :param ~pacman.model.placements.Placement placement:
            the placement to get the data from
        :param int recording_region_id: desired recording data region
        :return: an array contained all the data received during the
            simulation, and a flag indicating if any data was missing
        :rtype: tuple(bytearray, bool)
        """
        # Ensure that any transfers in progress are complete first
        if not isinstance(placement.vertex, AbstractReceiveBuffersToHost):
            raise NotImplementedError(
                "vertex {} does not implement AbstractReceiveBuffersToHost "
                "so no data read".format(placement.vertex))
        with self._thread_lock_buffer_out:
            # data flush has been completed - return appropriate data
            return self._received_data.get_region_data(
                placement.x, placement.y, placement.p, recording_region_id)

    def _retreive_by_placement(self, placement, region):
        """ Retrieve the data for a vertex; must be locked first.

        :param ~pacman.model.placements.Placement placement:
            the placement to get the data from
        :param int recording_region_id: desired recording data region
        """

        # Has the region information been read
        if not self._received_data.has_region_information(
                placement.x, placement.y, placement.p):

            addr = placement.vertex.get_recording_region_base_address(
                self._transceiver, placement)
            self._get_region_information(
                addr, placement.x, placement.y, placement.p)

        # Read the data if not already received
        if not self._received_data.is_data_from_region_flushed(
                placement.x, placement.y, placement.p, region):

            # Now read the data and store it
            size, addr, missing = self._received_data.get_region_information(
                placement.x, placement.y, placement.p, region)
            data = self._request_data(
                placement.x, placement.y, addr, size)
            self._received_data.store_data_in_region_buffer(
                placement.x, placement.y, placement.p, region, missing, data)

    def _get_region_information(self, addr, x, y, p):
        """ Get the recording information from all regions of a core

        :param addr: The recording region base address
        :param x: The x-coordinate of the chip containing the data
        :param y: The y-coordinate of the chip containing the data
        """
        n_regions = self._transceiver.read_word(x, y, addr)
        n_bytes = get_recording_header_size(n_regions)
        data = self._transceiver.read_memory(
            x, y, addr + BYTES_PER_WORD, n_bytes - BYTES_PER_WORD)
        data_type = _RecordingRegion * n_regions
        regions = data_type.from_buffer_copy(data)
        sizes_and_addresses = [
            (r.size, r.data, bool(r.missing)) for r in regions]
        self._received_data.store_region_information(
            x, y, p, sizes_and_addresses)

    @property
    def sender_vertices(self):
        """ The vertices which are buffered.

        :rtype: iterable(AbstractSendsBuffersFromHost)
        """
        return self._sender_vertices



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import defaultdict
from spinn_utilities.progress_bar import ProgressBar
from pacman.model.constraints.placer_constraints import ChipAndCoreConstraint
from spinn_front_end_common.utility_models import (
    LivePacketGather, LivePacketGatherMachineVertex)


class InsertLivePacketGatherersToGraphs(object):
    """ Adds LPGs as required into a given graph.
    """

    __slots__ = [
        "_machine_graph",
        "_application_graph"]

    def __call__(
            self, live_packet_gatherer_parameters, machine, machine_graph,
            application_graph=None):
        """ Add LPG vertices on Ethernet connected chips as required.

        :param live_packet_gatherer_parameters:
            the Live Packet Gatherer parameters requested by the script
        :type live_packet_gatherer_parameters:
            dict(LivePacketGatherParameters,
            list(tuple(~pacman.model.graphs.AbstractVertex, list(str))))
        :param ~spinn_machine.Machine machine:
            the SpiNNaker machine as discovered
        :param ~pacman.model.graphs.machine.MachineGraph machine_graph:
            the machine graph
        :param application_graph: the application graph
        :type application_graph:
            ~pacman.model.graphs.application.ApplicationGraph
        :return: mapping between LPG parameters and LPG vertex
        :rtype: dict(LivePacketGatherParameters,
            dict(tuple(int,int),LivePacketGatherMachineVertex))
        """

        self._machine_graph = machine_graph
        self._application_graph = application_graph

        # create progress bar
        progress = ProgressBar(
            machine.ethernet_connected_chips,
            string_describing_what_being_progressed=(
                "Adding Live Packet Gatherers to Graph"))

        # Keep track of the vertices added by parameters and board address
        lpg_params_to_vertices = defaultdict(dict)

        # for every Ethernet connected chip, add the gatherers required
        if application_graph is not None:
            for chip in progress.over(machine.ethernet_connected_chips):
                for params in live_packet_gatherer_parameters:
                    if (params.board_address is None or
                            params.board_address == chip.ip_address):
                        lpg_params_to_vertices[params][chip.x, chip.y] = \
                            self._add_app_lpg_vertex(chip, params)
        else:
            for chip in progress.over(machine.ethernet_connected_chips):
                for params in live_packet_gatherer_parameters:
                    if (params.board_address is None or
                            params.board_address == chip.ip_address):
                        lpg_params_to_vertices[params][chip.x, chip.y] = \
                            self._add_mach_lpg_vertex(chip, params)

        return lpg_params_to_vertices

    def _add_app_lpg_vertex(self, chip, params):
        """ Adds a LPG vertex to a machine graph that has an associated\
            application graph.

        :param ~.Chip chip:
        :param LivePacketGatherParameters params:
        :rtype: LivePacketGatherMachineVertex
        """
        app_vtx = LivePacketGather(params)
        self._application_graph.add_vertex(app_vtx)
        # No need to handle resources when allocating; LPG has core to itself
        vtx = app_vtx.create_machine_vertex(
            vertex_slice=None, resources_required=None,
            label="LivePacketGatherer",
            constraints=[ChipAndCoreConstraint(x=chip.x, y=chip.y)])
        self._machine_graph.add_vertex(vtx)
        return vtx

    def _add_mach_lpg_vertex(self, chip, params):
        """ Adds a LPG vertex to a machine graph without an associated\
            application graph.

        :param ~.Chip chip:
        :param LivePacketGatherParameters params:
        :rtype: LivePacketGatherMachineVertex
        """
        vtx = LivePacketGatherMachineVertex(
            params, constraints=[ChipAndCoreConstraint(x=chip.x, y=chip.y)])
        self._machine_graph.add_vertex(vtx)
        return vtx



# Copyright (c) 2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.abstract_base import (
    AbstractBase, abstractmethod)


class AbstractCanReset(object, metaclass=AbstractBase):
    """ Indicates an object that can be reset to time 0.

    This is used when AbstractSpinnakerBase.reset is called.
    All Vertices and all edges in the original graph
    (the one added to by the user) will be checked and reset.
    """
    __slots__ = []

    @abstractmethod
    def reset_to_first_timestep(self):
        """ Reset the object to first time step.
        """



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_front_end_common.utilities.utility_objs import ProvenanceDataItem


class PacmanProvenanceExtractor(object):
    """ Extracts Provenance data from a :py:class:`PACMANAlgorithmExecutor`
    """

    TOP_NAME = "pacman"

    __slots__ = ["__data_items", "__already_done"]

    def __init__(self):
        self.__data_items = list()
        self.__already_done = set()

    def extract_provenance(self, executor):
        """ Acquires the timings from PACMAN algorithms (provenance data)

        :param ~pacman.executor.PACMANAlgorithmExecutor executor:
            the PACMAN workflow executor
        :rtype: None
        """
        for (algorithm, run_time, exec_names) in executor.algorithm_timings:
            key = "run_time_of_{}".format(algorithm)
            if key not in self.__already_done:
                names = [self.TOP_NAME, exec_names, key]
                self.__data_items.append(ProvenanceDataItem(names, run_time))
                self.__already_done.add(key)

    @property
    def data_items(self):
        """ Returns the provenance data items

        :return: the provenance items
        :rtype: iterable(ProvenanceDataItem)
        """
        return self.__data_items

    def clear(self):
        """ Clears the provenance data store

        :rtype: None
        """
        self.__data_items = list()
        self.__already_done = set()



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Much of the code in this module is intended primarily for being invoked via the
PACMAN Executor.
"""

import os
from .bit_field_compressor_report import BitFieldCompressorReport
from .bit_field_summary import BitFieldSummary
from .board_chip_report import BoardChipReport
from .energy_report import EnergyReport
from .fixed_route_from_machine_report import FixedRouteFromMachineReport
from .memory_map_on_host_chip_report import MemoryMapOnHostChipReport
from .memory_map_on_host_report import MemoryMapOnHostReport
from .routing_table_from_machine_report import RoutingTableFromMachineReport
from .real_tags_report import TagsFromMachineReport


def report_xml():
    return os.path.join(
        os.path.dirname(__file__), "front_end_common_reports.xml")


__all__ = [
    "BitFieldCompressorReport",
    "BitFieldSummary",
    "BoardChipReport",
    "EnergyReport",
    "FixedRouteFromMachineReport",
    "MemoryMapOnHostChipReport",
    "MemoryMapOnHostReport",
    "report_xml",
    "RoutingTableFromMachineReport",
    "TagsFromMachineReport"]



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.progress_bar import ProgressBar
from pacman.model.resources import (
    SpecificChipSDRAMResource, CoreResource,
    PreAllocatedResourceContainer, SpecificBoardIPtagResource)
from spinn_front_end_common.utility_models import (
    ExtraMonitorSupportMachineVertex)
from spinn_front_end_common.utility_models import (
    DataSpeedUpPacketGatherMachineVertex)


class PreAllocateResourcesForExtraMonitorSupport(object):
    """ Allocates resources for the extra monitors.
    """
    def __call__(
            self, machine, pre_allocated_resources,
            n_cores_to_allocate=1):
        """
        :param ~spinn_machine.Machine machine: SpiNNaker machine object
        :param pre_allocated_resources: resources already preallocated
        :type pre_allocated_resources:
            ~pacman.model.resources.PreAllocatedResourceContainer
        :param int n_cores_to_allocate: how many gatherers to use per chip
        :rtype: ~pacman.model.resources.PreAllocatedResourceContainer
        """

        progress = ProgressBar(
            len(list(machine.ethernet_connected_chips)) + machine.n_chips,
            "Preallocating resources for Extra Monitor support vertices")

        sdrams = list()
        cores = list()
        tags = list()

        # add resource requirements for the gatherers on each Ethernet
        # connected chip. part of data extraction
        self._handle_packet_gathering_support(
            sdrams, cores, tags, machine, progress, n_cores_to_allocate)

        # add resource requirements for re-injector and reader for data
        # extractor
        self._handle_second_monitor_support(cores, sdrams, machine, progress)

        # note what has been preallocated
        allocated = PreAllocatedResourceContainer(
            specific_sdram_usage=sdrams, core_resources=cores,
            specific_iptag_resources=tags)
        allocated.extend(pre_allocated_resources)
        return allocated

    @staticmethod
    def _handle_second_monitor_support(cores, sdrams, machine, progress):
        """ Adds the second monitor preallocations, which reflect the\
            reinjector and data extractor support

        :param list(~.CoreResource) cores: the storage of core requirements
        :param ~.Machine machine: the spinnMachine instance
        :param ~.ProgressBar progress: the progress bar to operate one
        """
        extra_usage = \
            ExtraMonitorSupportMachineVertex.static_resources_required()
        for chip in progress.over(machine.chips):
            cores.append(CoreResource(chip=chip, n_cores=1))
            sdrams.append(SpecificChipSDRAMResource(
                chip=chip, sdram_usage=extra_usage.sdram))

    @staticmethod
    def _handle_packet_gathering_support(
            sdrams, cores, tags, machine, progress, n_cores_to_allocate):
        """ Adds the packet gathering functionality tied into the data\
            extractor within each chip

        :param list(~.SpecificChipSDRAMResource) sdrams:
            the preallocated SDRAM requirement for these vertices
        :param list(~.CoreResource) cores:
            the preallocated cores requirement for these vertices
        :param list(~.SpecificBoardIPtagResource) tags:
            the preallocated tags requirement for these vertices
        :param ~.Machine machine: the spinnMachine instance
        :param ~.ProgressBar progress: the progress bar to update as needed
        :param int n_cores_to_allocate:
            how many packet gathers to allocate per chip
        """
        # pylint: disable=too-many-arguments

        # get resources from packet gatherer
        resources = DataSpeedUpPacketGatherMachineVertex.\
            static_resources_required()

        # locate Ethernet connected chips that the vertices reside on
        for ethernet_connected_chip in \
                progress.over(machine.ethernet_connected_chips,
                              finish_at_end=False):
            # do resources. SDRAM, cores, tags
            sdrams.append(SpecificChipSDRAMResource(
                chip=ethernet_connected_chip,
                sdram_usage=resources.sdram))
            cores.append(CoreResource(
                chip=ethernet_connected_chip, n_cores=n_cores_to_allocate))
            tags.append(SpecificBoardIPtagResource(
                board=ethernet_connected_chip.ip_address,
                ip_address=resources.iptags[0].ip_address,
                port=resources.iptags[0].port,
                strip_sdp=resources.iptags[0].strip_sdp,
                tag=resources.iptags[0].tag,
                traffic_identifier=resources.iptags[0].traffic_identifier))



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .chip_power_monitor import ChipPowerMonitor
from .chip_power_monitor_machine_vertex import ChipPowerMonitorMachineVertex
from .command_sender import CommandSender
from .command_sender_machine_vertex import CommandSenderMachineVertex
from .data_speed_up_packet_gatherer import DataSpeedUpPacketGather
from .data_speed_up_packet_gatherer_machine_vertex import (
    DataSpeedUpPacketGatherMachineVertex)
from .extra_monitor_support import ExtraMonitorSupport
from .extra_monitor_support_machine_vertex import (
    ExtraMonitorSupportMachineVertex)
from .live_packet_gather import LivePacketGather
from .live_packet_gather_machine_vertex import LivePacketGatherMachineVertex
from .multi_cast_command import MultiCastCommand
from .reverse_ip_tag_multi_cast_source import ReverseIpTagMultiCastSource
from .reverse_ip_tag_multicast_source_machine_vertex import (
    ReverseIPTagMulticastSourceMachineVertex)

__all__ = ["CommandSender", "CommandSenderMachineVertex",
           "ChipPowerMonitor", "ChipPowerMonitorMachineVertex",
           "DataSpeedUpPacketGather", "DataSpeedUpPacketGatherMachineVertex",
           "ExtraMonitorSupport", "ExtraMonitorSupportMachineVertex",
           "LivePacketGather", "LivePacketGatherMachineVertex",
           "MultiCastCommand", "ReverseIpTagMultiCastSource",
           "ReverseIPTagMulticastSourceMachineVertex"]



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
from .sqllite_database import SqlLiteDatabase

#: Name of the database in the data folder
DB_FILE_NAME = "buffer.sqlite3"


class BufferedReceivingData(object):
    """ Stores the information received through the buffering output\
        from the SpiNNaker system.
    """

    __slots__ = [
        #: the AbstractDatabase holding the data to store
        "_db",

        #: the path to the database
        "_db_file",

        #: the (size, address) of each region
        "__sizes_and_addresses",

        #: whether data has been flushed for each region
        "__data_flushed"
    ]

    def __init__(self, report_folder):
        """
        :param str report_folder:
            The directory to write the database used to store some of the data
        """
        self._db_file = os.path.join(report_folder, DB_FILE_NAME)
        self._db = None
        self.reset()

    def reset(self):
        """ Perform tasks to restart recording from time=0
        """
        if os.path.exists(self._db_file):
            if self._db:
                self._db.close()
            os.remove(self._db_file)
        self._db = SqlLiteDatabase(self._db_file)
        self.__sizes_and_addresses = dict()
        self.__data_flushed = set()

    def resume(self):
        """ Perform tasks that will continue running without resetting
        """
        self.__sizes_and_addresses = dict()
        self.__data_flushed = set()

    def store_region_information(self, x, y, p, sizes_and_addresses):
        """ Store the sizes, addresses and is_missing data of the regions

        :param int x: The x-coordinate of the core whose data this is
        :param int y: The y-coordinate of the core whose data this is
        :param int p: The processor id of the core whose data this is
        :param list(int,int,bool) sizes_and_addresses:
            The size and address of each region
        """
        self.__sizes_and_addresses[x, y, p] = sizes_and_addresses

    def get_region_information(self, x, y, p, region_id):
        """ Get the size, address and is_missing of the region

        :param int x: The x-coordinate of the core whose data this is
        :param int y: The y-coordinate of the core whose data this is
        :param int p: The processor id of the core whose data this is
        :param int region_id: The id of the region to get the data for
        :rtype: tuple(int, int, bool)
        """
        return self.__sizes_and_addresses.get((x, y, p))[region_id]

    def has_region_information(self, x, y, p):
        """ Determine if region information has been stored for this core

        :param int x: The x-coordinate of the core whose data this is
        :param int y: The y-coordinate of the core whose data this is
        :param int p: The processor id of the core whose data this is
        :rtype: bool
        """
        return (x, y, p) in self.__sizes_and_addresses

    def store_data_in_region_buffer(self, x, y, p, region, missing, data):
        """ Store some information in the correspondent buffer class for a\
            specific chip, core and region.

        :param int x: x coordinate of the chip
        :param int y: y coordinate of the chip
        :param int p: Core within the specified chip
        :param int region: Region containing the data to be stored
        :param bool missing: Whether any data is missing
        :param bytearray data: data to be stored
        """
        # pylint: disable=too-many-arguments
        self._db.store_data_in_region_buffer(x, y, p, region, missing, data)
        self.__data_flushed.add((x, y, p, region))

    def is_data_from_region_flushed(self, x, y, p, region):
        """ Determine if data has been stored for this region

        :param int x: x coordinate of the chip
        :param int y: y coordinate of the chip
        :param int p: Core within the specified chip
        :param int region: Region containing the data
        """
        return (x, y, p, region) in self.__data_flushed

    def get_region_data(self, x, y, p, region):
        """ Get the data stored for a given region of a given core.

        :param int x: x coordinate of the chip
        :param int y: y coordinate of the chip
        :param int p: Core within the specified chip
        :param int region: Region containing the data
        :return: a buffer containing all the data received during the
            simulation, and a flag indicating if any data was missing
        :rtype: tuple(memoryview, bool)
        """
        return self._db.get_region_data(x, y, p, region)

    def clear(self, x, y, p, region_id):
        """ Clears the data from a given data region (only clears things\
            associated with a given data recording region).

        :param int x: placement x coordinate
        :param int y: placement y coordinate
        :param int p: placement p coordinate
        :param int region_id: the recording region ID to clear data from
        :rtype: None
        """
        self._db.clear_region(x, y, p, region_id)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from spinn_utilities.executable_finder import ExecutableFinder as BaseEF
from spinn_front_end_common import common_model_binaries


class ExecutableFinder(BaseEF):
    """ Manages a set of folders in which to search for binaries, and allows\
        for binaries to be discovered within this path. This adds a default\
        location to look to the base class.
    """

    def __init__(self, binary_search_paths=None,
                 include_common_binaries_folder=True):
        """
        :param iterable(str) binary_search_paths:
            The initial set of folders to search for binaries.
        :param bool include_common_binaries_folder:
            If True (i.e. the default), the
            spinn_front_end_common.common_model_binaries folder is searched
            for binaries.  If you are not using the common models, or the
            model binary names conflict with your own, this parameter can be
            used to avoid this issue. Note that the folder will be appended
            to the value of binary_search_paths, so the common binary search
            path will be looked in last.
        """
        if binary_search_paths is None:
            binary_search_paths = list()
        super().__init__(binary_search_paths)
        if include_common_binaries_folder:
            self.add_path(os.path.dirname(common_model_binaries.__file__))



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import struct
import logging
from spinn_utilities.progress_bar import ProgressBar
from spinn_utilities.log import FormatAdapter
from spinnman.messages.sdp import SDPFlag, SDPHeader, SDPMessage
from spinnman.model.enums import CPUState
from spinn_front_end_common.utilities.constants import (
    SDP_PORTS, SDP_RUNNING_MESSAGE_CODES)
from spinn_front_end_common.utilities.exceptions import ConfigurationException

logger = FormatAdapter(logging.getLogger(__name__))
_ONE_WORD = struct.Struct("<I")

_LIMIT = 10


class ChipProvenanceUpdater(object):
    """ Forces all cores to generate provenance data, and then exit.
    """

    __slots__ = []

    def __call__(self, txrx, app_id, all_core_subsets):
        """
        :param ~spinnman.transceiver.Transceiver txrx:
        :param int app_id:
        :param ~spinn_machine.CoreSubsets all_core_subsets:
        """
        # check that the right number of processors are in sync
        processors_completed = txrx.get_core_state_count(
            app_id, CPUState.FINISHED)
        total_processors = len(all_core_subsets)
        left_to_do_cores = total_processors - processors_completed

        progress = ProgressBar(
            left_to_do_cores,
            "Forcing error cores to generate provenance data")

        error_cores = txrx.get_cores_in_state(
            all_core_subsets, CPUState.RUN_TIME_EXCEPTION)
        watchdog_cores = txrx.get_cores_in_state(
            all_core_subsets, CPUState.WATCHDOG)
        idle_cores = txrx.get_cores_in_state(
            all_core_subsets, CPUState.IDLE)

        if error_cores or watchdog_cores or idle_cores:
            raise ConfigurationException(
                "Some cores have crashed. RTE cores {}, watch-dogged cores {},"
                " idle cores {}".format(
                    error_cores.values(), watchdog_cores.values(),
                    idle_cores.values()))

        # check that all cores are in the state FINISHED which shows that
        # the core has received the message and done provenance updating
        self._update_provenance(txrx, total_processors, processors_completed,
                                all_core_subsets, app_id, progress)
        progress.end()

    def _update_provenance(self, txrx, total_processors, processors_completed,
                           all_core_subsets, app_id, progress):
        """
        :param ~.Transceiver txrx:
        :param int total_processors:
        :param int processors_completed:
        :param ~.CoreSubsets all_core_subsets:
        :param int app_id:
        :param ~.ProgressBar progress:
        """
        # pylint: disable=too-many-arguments
        left_to_do_cores = total_processors - processors_completed
        attempts = 0
        while processors_completed != total_processors and attempts < _LIMIT:
            attempts += 1
            unsuccessful_cores = txrx.get_cores_not_in_state(
                all_core_subsets, CPUState.FINISHED)

            for (x, y, p) in unsuccessful_cores.keys():
                self._send_chip_update_provenance_and_exit(txrx, x, y, p)

            processors_completed = txrx.get_core_state_count(
                app_id, CPUState.FINISHED)

            left_over_now = total_processors - processors_completed
            to_update = left_to_do_cores - left_over_now
            left_to_do_cores = left_over_now
            if to_update != 0:
                progress.update(to_update)
        if processors_completed != total_processors:
            logger.error("Unable to Finish getting provenance data. "
                         "Abandoned after too many retries. "
                         "Board may be left in an unstable state!")

    @staticmethod
    def _send_chip_update_provenance_and_exit(txrx, x, y, p):
        """
        :param ~.Transceiver txrx:
        :param int x:
        :param int y:
        :param int p:
        """
        cmd = SDP_RUNNING_MESSAGE_CODES.SDP_UPDATE_PROVENCE_REGION_AND_EXIT
        port = SDP_PORTS.RUNNING_COMMAND_SDP_PORT

        txrx.send_sdp_message(SDPMessage(
            SDPHeader(
                flags=SDPFlag.REPLY_NOT_EXPECTED,
                destination_port=port.value, destination_cpu=p,
                destination_chip_x=x, destination_chip_y=y),
            data=_ONE_WORD.pack(cmd.value)))



# Copyright (c) 2019-2020 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_utilities.require_subclass import require_subclass
from pacman.model.graphs.machine import MachineVertex


@require_subclass(MachineVertex)
class AbstractSupportsBitFieldRoutingCompression(
        object, metaclass=AbstractBase):
    """ Marks a machine vertex that can support having the on-chip bitfield \
        compressor running on its core.
    """
    __slots__ = ()

    @abstractmethod
    def key_to_atom_map_region_base_address(self, transceiver, placement):
        """ Returns the SDRAM address for the region that contains \
            key-to-atom data.

        :param ~spinnman.transceiver.Transceiver transceiver:
        :param ~pacman.model.placements.Placement placement:
        :return: the SDRAM address for the key-to-atom data
        :rtype: int
        """

    @abstractmethod
    def bit_field_base_address(self, transceiver, placement):
        """ Returns the SDRAM address for the bit-field table data.

        :param ~spinnman.transceiver.Transceiver transceiver:
        :param ~pacman.model.placements.Placement placement:
        :return: the SDRAM address for the bitfield address
        :rtype: int
        """

    @abstractmethod
    def regeneratable_sdram_blocks_and_sizes(self, transceiver, placement):
        """ Returns the SDRAM addresses and sizes for the cores' SDRAM that \
            are available (borrowed) for generating bitfield tables.

        :param ~spinnman.transceiver.Transceiver transceiver:
        :param ~pacman.model.placements.Placement placement:
        :return: list of tuples containing (the SDRAM address for the cores
            SDRAM address's for the core's SDRAM that can be used to generate
            bitfield tables loaded, and the size of memory chunks located
            there)
        :rtype: list(tuple(int,int))
        """
